from abc import ABCMeta, abstractmethod, abstractproperty
import logging

try:
    from collections.abc import Iterable
except ImportError:
    from collections import Iterable

import numpy as np


try:
    import scipy
    import scipy.interpolate
    import scipy.linalg

    SCIPY_FLAG = True
except Exception:
    SCIPY_FLAG = False

from .fourier_fitting import FourierFit


log = logging.getLogger(__name__)


class WESTStringMethod:
    ___metaclass__ = ABCMeta

    def __init__(self, centers, **kwargs):
        pass

    @abstractproperty
    def centers(self):
        """Return the centers of all of the strings"""
        pass

    @abstractproperty
    def length(self):
        """Return a list of the lengths of each string"""
        pass

    @abstractmethod
    def update_string_centers(self, avgcoords, binprob):
        """Given a set of average coordinates (avgcoords) in each bin
        and the individual probabilities for each bin (binprob), update
        the string centers
        """
        pass


class DefaultStringMethod(WESTStringMethod):
    """Implementation of a method to evolve one or more pseudo-1D strings in a high dimensional
    progress coordinate space.

    **Parameters**
    centers:        A numpy array of size (number of total centers,pcoord dim) that stores
                        the positions of all of the string images
    slen:           An iterable containing the number of centers in each string
    slabels:        An list containing the relative positions in each string of any state label
                        progress coordinates if present. These progress coordinates will be ignored in the
                        calculation. None if no labels
    mpairs:         A list of lists containing the indices of pairs of centers that should move together.
                        None if strings move independently
    dtau:           Parameter controlling the rate at which centers move toward the average value in the bin
    kappa:          Parameter controlling the smoothing of the string
    fixed_ends:     Boolean flag specifying whether to fix ends of the strings
    sciflag:        Boolean flag specifying whether to attempt to use scipy methods which are
                        generally more efficient
    fourierflag:    Boolean flag specifying whether to user fourier fitting method
    fourier_P:      Integer value specifying how many fourier modes to use in fitting
    fourier_maxiters: Maximum number of iterations of fourier fitting procedure
    fourier_tol:    Tolerance for ending fourier fitting
    """

    def __init__(
        self,
        centers,
        slen=None,
        slabels=None,
        mpairs=None,
        dtau=0.1,
        kappa=0.1,
        sciflag=None,
        fixed_ends=True,
        fourierflag=False,
        fourier_P=2,
        fourier_maxiters=100,
        fourier_tol=1.0e-6,
        **kwargs
    ):
        super().__init__(centers, **kwargs)

        self._SCIPY_FLAG = None
        self._fixed_ends = fixed_ends

        if sciflag is None:
            self._SCIPY_FLAG = SCIPY_FLAG
        else:
            self._SCIPY_FLAG = SCIPY_FLAG & sciflag

        self._dtau = dtau
        self._kappa = kappa

        self._centers = centers

        self._nstrings = len(slen)  # Number of strings
        self._N = centers.shape[0]
        self._ndim = centers.shape[1]  # Number of progress coordinates

        if slen and isinstance(slen, Iterable):
            self._slen = np.array(slen)
        elif slen:
            self._slen = np.array([slen])
        else:
            self._slen = np.array([self._N])
            log.warning('Input parameters do not define slen; assuming system is composed of a single string')

        self._mpairs = mpairs

        assert np.sum(self._slen) == self._N

        self._skip_dim = np.array(slabels) if slabels is not None else np.array([])

        # Get iterable of slicing objects to get per string center coordinates to perform calculation on
        self._strindx = []

        indx_all = np.arange(self._ndim)
        self._indx_take = np.setdiff1d(indx_all, self._skip_dim)
        self._ndim_take = len(self._indx_take)

        start_count = 0
        for sl in slen:
            self._strindx.append(np.index_exp[start_count : (start_count + sl), self._indx_take])
            start_count += sl

        # Fourier fitting parameters
        self._FFIT_FLAG = fourierflag
        self._ffp = fourier_P
        self._ffmaxiters = fourier_maxiters
        self._fftol = fourier_tol

        # Create dict to hold kappan and A objects for all unique lengths of strings
        self._kappan = {}
        self._A = {}

        self.finalize_init()

    @property
    def centers(self):
        return self._centers

    @property
    def length(self):
        L = []
        for sid, si in enumerate(self._strindx):
            L.append(self.calculate_length(self._centers[si])[-1])

        return L

    def calculate_length(self, x):
        dd = x - np.roll(x, 1, axis=0)
        dd[0, :] = 0.0
        return np.cumsum(np.sqrt((dd * dd).sum(axis=1)))

    def finalize_init(self):
        # Set up A and kappan for each string
        uslen = np.unique(self._slen)

        for ulen in uslen:
            self._kappan[ulen] = self._kappa * self._dtau * ulen
            self._A[ulen] = None

            if self._SCIPY_FLAG:
                ud = np.zeros((ulen,))
                ld = np.zeros((ulen,))
                d = np.ones((ulen,))

                d[1:-1] = 2.0 * self._kappan[ulen] + 1.0
                ud[2:] = -self._kappan[ulen]
                ld[:-2] = -self._kappan[ulen]

                self._A[ulen] = np.mat([ud, d, ld])

            else:
                self._A[ulen] = np.eye(ulen)
                di = np.diag_indices(ulen, ndim=2)
                ii = (di[0][1:-1], di[1][1:-1])

                self._A[ulen][ii] = 2.0 * self._kappan[ulen] + 1.0

                dd = np.zeros((ulen - 1,))
                dd[1:] = -self._kappan[ulen]
                self._A[ulen] += np.diag(dd, k=1)

                dd = np.zeros((ulen - 1,))
                dd[:-1] = -self._kappan[ulen]
                self._A[ulen] += np.diag(dd, k=-1)

    def update_string_centers(self, avgcoords, binprob):
        """Update the position of all string centers
        **Parameters**
        avgcoords:      Average position of replicas in each voronoi cell
        binprob:        The total weight in each voronoi cell

        """
        assert self.centers.shape == avgcoords.shape

        # If centers are paired, calculate their average position
        if self._mpairs is not None:
            for pi in self._mpairs:
                pprob = binprob[pi]
                if np.sum(pprob) == 0.0:
                    continue

                idx = np.ix_(pi, self._indx_take)
                pavg = np.ma.array(avgcoords[idx])
                zind = np.where(pprob == 0)[0]
                pavg[zind, :] = np.ma.masked

                avgcoords[idx] = pavg.mean(axis=0)

        for sid, si in enumerate(self._strindx):
            x = avgcoords.copy()[si]
            centers = self.centers[si]
            occupied = np.nonzero(binprob[si[0]])

            N = self._slen[sid]

            # if avgcoords has missing values fill them by linearly interpolating
            # present data
            if occupied[0].shape != N:
                notocc = np.ones((N,), dtype=bool)  # unoccupied
                notocc[occupied] = False

                # marked paired centers as occupied to avoid reseting averaged value if
                # at least one is occupied
                if self._mpairs is not None:
                    for pi in self._mpairs:
                        totprob = np.sum(binprob[pi])
                        if totprob == 0.0:
                            continue
                        else:
                            for m in pi:
                                if (m >= si[0].start) and (m < si[0].stop):
                                    notocc[m - si[0].start] = False

                cfunc = lambda z: z.nonzero()[0]

                # Handle ends first
                if notocc[0]:
                    x[0, :] = centers[0, :]
                    notocc[0] = False
                if notocc[-1]:
                    x[-1, :] = centers[-1, :]
                    notocc[-1] = False

                # interpolate values for unoccupied bins
                if self._SCIPY_FLAG:
                    for k in range(self._ndim_take):
                        f = scipy.interpolate.interp1d(cfunc(~notocc), x[~notocc, k], kind='linear')
                        x[notocc, k] = f(cfunc(notocc))
                else:
                    for k in range(self._ndim_take):
                        x[notocc, k] = np.interp(cfunc(notocc), cfunc(~notocc), x[~notocc, k])

            if self._fixed_ends:
                x[0, :] = centers[0, :]
                x[-1, :] = centers[-1, :]

            psi = centers
            psi_new = np.zeros_like(psi)

            b = psi - self._dtau * (psi - x)

            # Update and smooth the string
            if self._SCIPY_FLAG:
                for k in range(self._ndim_take):
                    psi_new[:, k] = scipy.linalg.solve_banded((1, 1), self._A[N], b[:, k])
            else:
                for k in range(self._ndim_take):
                    psi_new[:, k] = np.linalg.solve(self._A[N], b[:, k])

            # Optionally smooth using fourier method
            if self._FFIT_FLAG:
                w0 = np.zeros((self._ndim_take, self._ffp), np.float)
                t0 = np.linspace(0, 1, psi_new.shape[0])

                ff = FourierFit(P=self._ffp, maxiters=self._ffmaxiters)
                ff.optimize(psi_new, None, w0, t0)
                psi_new = ff.pp[-1][:]

            # Enforce equal spacing between centers along the string
            L = self.calculate_length(psi_new)
            L /= L[-1]
            g2 = np.linspace(0, 1, N)

            if self._SCIPY_FLAG:
                for k in range(self._ndim_take):
                    f = scipy.interpolate.interp1d(L, psi_new[:, k], kind='linear')
                    psi_new[:, k] = f(g2)
            else:
                for k in range(self._ndim_take):
                    psi_new[:, k] = np.interp(g2, L, psi_new[:, k])

            self.centers[si] = psi_new.copy()



from westpa.analysis.core import Run
from westpa.analysis.statistics import time_average
from westpa.analysis.trajectories import Trajectory, BasicMDTrajectory, HDF5MDTrajectory

__all__ = ['Run', 'time_average', 'Trajectory', 'BasicMDTrajectory', 'HDF5MDTrajectory']



import logging
from argparse import ArgumentParser

import westpa


log = logging.getLogger('w_truncate')

warning_string = '''\
NOTE: w_truncate only deletes iteration groups from the HDF5 data store.
It is recommended that any iteration data saved to the file system (e.g. in the
traj_segs directory) is deleted or moved for the corresponding iterations.
'''


def entry_point():
    parser = ArgumentParser(
        'w_truncate',
        description='''\
    Remove all iterations after a certain point in a WESTPA simulation.
    ''',
        epilog=warning_string,
    )

    westpa.rc.add_args(parser)
    parser.add_argument(
        '-W',
        '--west-data',
        dest='we_h5filename',
        metavar='WEST_H5FILE',
        help='''Take WEST data from WEST_H5FILE (default: read from the HDF5 file specified in west.cfg).''',
    )
    parser.add_argument('-n', '--iter', dest='n_iter', type=int, default=0, help='Truncate this iteration and those following.')

    args = parser.parse_args()
    westpa.rc.process_args(args, config_required=False)
    dm = westpa.rc.get_data_manager()
    if args.we_h5filename:
        dm.we_h5filename = args.we_h5filename

    dm.open_backing()
    # max_iter = dm.current_iteration

    if args.n_iter > dm.current_iteration:
        parser.error(
            'Provided iteration {} > current iteration {} of the {} HDF5 file. Exiting without doing anything.'.format(
                args.n_iter, dm.current_iteration, dm.we_h5filename.split('/')[-1]
            )
        )
    else:
        n_iter = args.n_iter if args.n_iter > 0 else dm.current_iteration

    for i in range(n_iter, dm.current_iteration + 1):
        dm.del_iter_group(i)

    dm.del_iter_summary(n_iter)
    dm.current_iteration = n_iter - 1

    westpa.rc.pstatus('simulation data truncated after iteration {}'.format(dm.current_iteration))
    westpa.rc.pstatus('\n' + warning_string)
    westpa.rc.pflush()

    dm.flush_backing()
    dm.close_backing()


if __name__ == '__main__':
    entry_point()



import logging

import numpy as np

log = logging.getLogger(__name__)

EPS = np.finfo(np.float64).eps


class Bin(set):
    def __init__(self, iterable=None, label=None):
        super().__init__(iterable or [])
        self.label = label

    def __repr__(self):
        return '<{classname} at 0x{id:x}, label={label!r}, count={count:d}, weight={weight:g}>'.format(
            classname=self.__class__.__name__, id=id(self), label=self.label, count=len(self), weight=self.weight
        )

    @property
    def weight(self):
        'Total weight of all walkers in this bin'

        weight = 0.0
        for particle in self:
            weight += particle.weight
        return weight

    def reweight(self, new_weight):
        """Reweight all walkers in this bin so that the total weight is new_weight"""

        if len(self) == 0 and new_weight == 0:
            return

        if len(self) == 0 and new_weight != 0:
            raise ValueError('cannot reweight empty ParticleCollection')

        current_weight = self.weight
        log.debug('reweighting collection of {:d} particles from {:g} to {:g}'.format(len(self), current_weight, new_weight))
        assert (new_weight == 0 and current_weight == 0) or new_weight > 0

        wrat = new_weight / current_weight
        for p in self:
            p.weight *= wrat

        log.debug('new weight: {:g}'.format(self.weight))
        assert abs(new_weight - self.weight) <= EPS * len(self)



if __name__ == '__main__':
    from . import autocorrel_elem
    import numpy
    from scipy.signal import correlate

    n = 16
    x = numpy.linspace(0, n * numpy.pi, 16 * n + 1)
    a = numpy.cos(x) + numpy.exp(-((x / 2.0) ** 2)) + numpy.exp(-(x / 4.0))
    pa = numpy.zeros((10000 * len(a),), numpy.float64)
    pa[: len(a)] = a

    print('<a> =', a.mean())
    print('<a^2> =', ((a - a.mean()) ** 2).sum())
    print('scipy.signal.correlate:')
    acf0 = correlate(a, a)
    acf0 = acf0[-len(a) :]
    acf0 /= acf0.max()
    print(acf0[: len(acf0) / 4])

    #    print 'scipy.signal.correlate (-mean):'
    #    acf0 = correlate(a-a.mean(),a-a.mean())
    #    acf0 = acf0[-len(a):]
    #    acf0 /= acf0.max()
    #    print acf0[:len(acf0)/4]

    print('this module:')
    acf = numpy.array([autocorrel_elem(pa, k) for k in range(len(a))])
    print(acf[: len(acf) / 4])



from westpa.tools import WESTParallelTool, WESTDataReader, IterRangeSelection, ProgressIndicatorComponent

import numpy as np

from westpa.core import h5io
from westpa.core.data_manager import seg_id_dtype, n_iter_dtype, weight_dtype
from westpa.core.extloader import get_object


def _find_matching_segments(west_datafile_name, n_iter, predicate, invert=False):
    '''Find all segments in iteration ``n_iter`` that match (or do not match, if
    ``invert`` is true) the given ``predicate``. Returns a sequence of matching
    seg_ids.'''

    with h5io.WESTPAH5File(west_datafile_name, 'r') as west_datafile:
        iter_group = west_datafile.get_iter_group(n_iter)
        nsegs = iter_group['seg_index'].shape[0]
        matching_ids = set(map(int, predicate(n_iter, iter_group)))

        if invert:
            matching_ids = set(range(nsegs)) - matching_ids

        matchvec = sorted(np.fromiter(matching_ids, dtype=seg_id_dtype, count=len(matching_ids)))
        return n_iter, matchvec


class WSelectTool(WESTParallelTool):
    prog = 'w_select'
    description = '''\
Select dynamics segments matching various criteria. This requires a
user-provided prediate function. By default, only matching segments are
stored. If the -a/--include-ancestors option is given, then matching segments
and their ancestors will be stored.


-----------------------------------------------------------------------------
Predicate function
-----------------------------------------------------------------------------

Segments are selected based on a predicate function, which must be callable
as ``predicate(n_iter, iter_group)`` and return a collection of segment IDs
matching the predicate in that iteration.

The predicate may be inverted by specifying the -v/--invert command-line
argument.


-----------------------------------------------------------------------------
Output format
-----------------------------------------------------------------------------

The output file (-o/--output, by default "select.h5") contains the following
datasets:

  ``/n_iter`` [iteration]
    *(Integer)* Iteration numbers for each entry in other datasets.

  ``/n_segs`` [iteration]
    *(Integer)* Number of segment IDs matching the predicate (or inverted
    predicate, if -v/--invert is specified) in the given iteration.

  ``/seg_ids`` [iteration][segment]
    *(Integer)* Matching segments in each iteration. For an iteration
    ``n_iter``, only the first ``n_iter`` entries are valid. For example,
    the full list of matching seg_ids in the first stored iteration is
    ``seg_ids[0][:n_segs[0]]``.

  ``/weights`` [iteration][segment]
    *(Floating-point)* Weights for each matching segment in ``/seg_ids``.


-----------------------------------------------------------------------------
Command-line arguments
-----------------------------------------------------------------------------
'''

    def __init__(self):
        super().__init__()

        self.data_reader = WESTDataReader()
        self.iter_range = IterRangeSelection()
        self.progress = ProgressIndicatorComponent()
        self.output_file = None
        self.output_filename = None
        self.predicate = None
        self.invert = False
        self.include_ancestors = False

    def add_args(self, parser):
        self.data_reader.add_args(parser)
        self.iter_range.add_args(parser)

        sgroup = parser.add_argument_group('selection options')
        sgroup.add_argument(
            '-p',
            '--predicate-function',
            metavar='MODULE.FUNCTION',
            help='''Use the given predicate function to match segments. This function
                             should take an iteration number and the HDF5 group corresponding to that
                             iteration and return a sequence of seg_ids matching the predicate, as in
                             ``match_predicate(n_iter, iter_group)``.''',
        )
        sgroup.add_argument('-v', '--invert', dest='invert', action='store_true', help='''Invert the match predicate.''')
        sgroup.add_argument(
            '-a', '--include-ancestors', action='store_true', help='''Include ancestors of matched segments in output.'''
        )

        ogroup = parser.add_argument_group('output options')
        ogroup.add_argument('-o', '--output', default='select.h5', help='''Write output to OUTPUT (default: %(default)s).''')
        self.progress.add_args(parser)

    def process_args(self, args):
        self.progress.process_args(args)
        self.data_reader.process_args(args)
        with self.data_reader:
            self.iter_range.process_args(args)

        predicate = get_object(args.predicate_function, path=['.'])
        if not callable(predicate):
            raise TypeError('predicate object {!r} is not callable'.format(predicate))
        self.predicate = predicate
        self.invert = bool(args.invert)
        self.include_ancestors = bool(args.include_ancestors)
        self.output_filename = args.output

    def go(self):
        self.data_reader.open('r')
        output_file = h5io.WESTPAH5File(self.output_filename, mode='w')
        pi = self.progress.indicator

        iter_start, iter_stop = self.iter_range.iter_start, self.iter_range.iter_stop
        iter_count = iter_stop - iter_start

        output_file.create_dataset('n_iter', dtype=n_iter_dtype, data=list(range(iter_start, iter_stop)))
        current_seg_count = 0
        seg_count_ds = output_file.create_dataset('n_segs', dtype=np.uint, shape=(iter_count,))
        matching_segs_ds = output_file.create_dataset(
            'seg_ids',
            shape=(iter_count, 0),
            maxshape=(iter_count, None),
            dtype=seg_id_dtype,
            chunks=h5io.calc_chunksize((iter_count, 1000000), seg_id_dtype),
            shuffle=True,
            compression=9,
        )
        weights_ds = output_file.create_dataset(
            'weights',
            shape=(iter_count, 0),
            maxshape=(iter_count, None),
            dtype=weight_dtype,
            chunks=h5io.calc_chunksize((iter_count, 1000000), weight_dtype),
            shuffle=True,
            compression=9,
        )

        with pi:
            pi.new_operation('Finding matching segments', extent=iter_count)
            #             futures = set()
            #             for n_iter in xrange(iter_start,iter_stop):
            #                 futures.add(self.work_manager.submit(_find_matching_segments,
            #                                                      args=(self.data_reader.we_h5filename,n_iter,self.predicate,self.invert)))

            #             for future in self.work_manager.as_completed(futures):
            for future in self.work_manager.submit_as_completed(
                (
                    (_find_matching_segments, (self.data_reader.we_h5filename, n_iter, self.predicate, self.invert), {})
                    for n_iter in range(iter_start, iter_stop)
                ),
                self.max_queue_len,
            ):
                n_iter, matching_ids = future.get_result()
                n_matches = len(matching_ids)

                if n_matches:
                    if n_matches > current_seg_count:
                        current_seg_count = len(matching_ids)
                        matching_segs_ds.resize((iter_count, n_matches))
                        weights_ds.resize((iter_count, n_matches))
                        current_seg_count = n_matches

                    seg_count_ds[n_iter - iter_start] = n_matches
                    matching_segs_ds[n_iter - iter_start, :n_matches] = matching_ids
                    weights_ds[n_iter - iter_start, :n_matches] = self.data_reader.get_iter_group(n_iter)['seg_index']['weight'][
                        sorted(matching_ids)
                    ]
                del matching_ids
                pi.progress += 1

            if self.include_ancestors:
                pi.new_operation('Tracing ancestors of matching segments', extent=iter_count)
                from_previous = set()
                current_seg_count = matching_segs_ds.shape[1]
                for n_iter in range(iter_stop - 1, iter_start - 1, -1):
                    iiter = n_iter - iter_start
                    n_matches = seg_count_ds[iiter]
                    matching_ids = set(from_previous)
                    if n_matches:
                        matching_ids.update(matching_segs_ds[iiter, : seg_count_ds[iiter]])
                    from_previous.clear()

                    n_matches = len(matching_ids)
                    if n_matches > current_seg_count:
                        matching_segs_ds.resize((iter_count, n_matches))
                        weights_ds.resize((iter_count, n_matches))
                        current_seg_count = n_matches

                    if n_matches > 0:
                        seg_count_ds[iiter] = n_matches
                        matching_ids = sorted(matching_ids)
                        matching_segs_ds[iiter, :n_matches] = matching_ids
                        weights_ds[iiter, :n_matches] = self.data_reader.get_iter_group(n_iter)['seg_index']['weight'][
                            sorted(matching_ids)
                        ]
                        parent_ids = self.data_reader.get_iter_group(n_iter)['seg_index']['parent_id'][sorted(matching_ids)]
                        from_previous.update(parent_id for parent_id in parent_ids if parent_id >= 0)  # filter initial states
                        del parent_ids
                    del matching_ids
                    pi.progress += 1


def entry_point():
    WSelectTool().main()


if __name__ == '__main__':
    entry_point()



from ._fasthist import histnd  # noqa

import numpy as np


def normhistnd(hist, binbounds):
    '''Normalize the N-dimensional histogram ``hist`` with corresponding
    bin boundaries ``binbounds``.  Modifies ``hist`` in place and returns
    the normalization factor used.'''

    ndim = hist.ndim

    if ndim != len(binbounds):
        raise ValueError(
            'shape of histogram [{!r}] does not match bin boundary sets (there are {})'.format(hist.shape, len(binbounds))
        )

    diffs = [np.diff(bb) for bb in binbounds]

    if ndim == 1:
        assert diffs[0].shape == hist.shape
        normfac = (hist * diffs[0]).sum()
    else:
        outers = np.multiply.outer(diffs[0], diffs[1])
        for delta in diffs[2:]:
            outers = np.multiply.outer(outers, delta)
        assert outers.shape == hist.shape, 'hist shape {} != outers shape {}'.format(hist.shape, outers.shape)
        # Divide by bin volumes
        hist /= outers
        normfac = hist.sum()
        # normfac = (hist * outers).sum()

    hist /= normfac
    return normfac



import pytest
from unittest import TestCase

import numpy as np

from westpa.core.binning import RectilinearBinMapper
from westpa.core.segment import Segment
from westpa.core.states import TargetState, InitialState
from westpa.core.systems import WESTSystem
from westpa.core.we_driver import WEDriver

EPS = np.finfo(np.float64).eps


class TestWEDriver(TestCase):
    def setUp(self):
        system = WESTSystem()
        system.bin_mapper = RectilinearBinMapper([[0.0, 1.0, 2.0]])
        system.bin_target_counts = np.array([4, 4])
        system.pcoord_len = 2
        self.we_driver = WEDriver(system=system)
        self.system = system
        self._seg_id = 0

    def segment(self, init_pcoord, final_pcoord, weight=1.0):
        segment = Segment(n_iter=1, seg_id=self._seg_id, pcoord=self.system.new_pcoord_array(), weight=weight)
        segment.pcoord[0] = init_pcoord
        segment.pcoord[1] = final_pcoord
        self._seg_id += 1
        return segment

    def tearDown(self):
        self.we_driver.clear()
        del self.we_driver
        del self.system
        del self._seg_id

    def test_assign(self):
        segments = [self.segment(0.0, 1.5, weight=0.5), self.segment(1.5, 0.5, weight=0.5)]
        self.we_driver.new_iteration()
        n_recycled = self.we_driver.assign(segments)
        assert n_recycled == 0
        assert len(self.we_driver.initial_binning[0]) == 1
        assert len(self.we_driver.initial_binning[1]) == 1
        assert len(self.we_driver.final_binning[0]) == 1
        assert len(self.we_driver.final_binning[1]) == 1
        assert (self.we_driver.flux_matrix == np.array([[0.0, 0.5], [0.5, 0.0]])).all()

    def test_passthrough(self):
        segments = [self.segment(0.0, 1.5, weight=0.125) for _i in range(4)] + [
            self.segment(1.5, 0.5, weight=0.125) for _i in range(4)
        ]
        segs_by_id = {segment.seg_id: segment for segment in segments}
        self.we_driver.new_iteration()
        self.we_driver.assign(segments)
        self.we_driver.construct_next()
        out_segs = set()
        for bin in self.we_driver.next_iter_binning:
            out_segs.update(bin)
        assert out_segs == set(self.we_driver.next_iter_segments)
        assert abs(sum(seg.weight for seg in self.we_driver.next_iter_segments) - 1.0) < 8 * EPS

        for segment in self.we_driver.next_iter_segments:
            # has n_iter been advanced?
            assert segment.n_iter == 2

            # is weight correct?
            assert segment.weight == 0.125  # rigorous floating point comparison okay here because no math was done

            # is parent set correctly?
            assert segment.parent_id is not None

            # was pcoord set correctly?
            assert (segs_by_id[segment.parent_id].pcoord[-1] == segment.pcoord[0]).all()

            # was status set correctly?
            assert segment.status == Segment.SEG_STATUS_PREPARED

        # were parent endpoint types set correctly
        for segment in segments:
            assert segment.endpoint_type == Segment.SEG_ENDPOINT_CONTINUES

    def test_split_by_weight(self):
        segments = [self.segment(1.5, 0.5, weight=0.25), self.segment(0.0, 1.5, weight=0.75)]
        self.we_driver.new_iteration()
        self.we_driver.assign(segments)
        self.we_driver.construct_next()
        assert len(self.we_driver.next_iter_binning[0]) == 4
        assert len(self.we_driver.next_iter_binning[1]) == 4
        assert abs(sum(seg.weight for seg in self.we_driver.next_iter_binning[0]) - 0.25) < 4 * EPS
        assert abs(sum(seg.weight for seg in self.we_driver.next_iter_binning[1]) - 0.75) < 4 * EPS
        assert np.allclose([seg.weight for seg in self.we_driver.next_iter_binning[0]], [0.25 / 4.0 for _i in range(4)])
        assert np.allclose([seg.weight for seg in self.we_driver.next_iter_binning[1]], [0.75 / 4.0 for _i in range(4)])

        for ibin in range(2):
            for segment in self.we_driver.next_iter_binning[ibin]:
                print(segment)
                assert segment.n_iter == 2
                assert segment.parent_id is not None
                assert segment.parent_id == segments[ibin].seg_id
                assert segment.status == Segment.SEG_STATUS_PREPARED

    # this test will fail up to alpha of the time
    @pytest.mark.flaky(reruns=5, only_rerun="AssertionError")
    def test_merge_by_weight(self):
        selected_counts = {0: 0, 1: 0}
        alpha = 0.01
        nrounds = 1000
        from scipy.stats import binom

        # lower and upper bounds of 95% CI for selecting the segment with weight 1/3
        lb = binom.ppf(alpha / 2.0, nrounds, 1.0 / 3.0)
        ub = binom.ppf(1.0 - alpha / 2.0, nrounds, 1.0 / 3.0)

        system = WESTSystem()
        system.bin_mapper = RectilinearBinMapper([[0.0, 1.0]])
        system.bin_target_counts = np.array([1])
        system.pcoord_len = 2
        self.we_driver = WEDriver(system=system)
        self.system = system
        self._seg_id = 0

        segments = [
            Segment(n_iter=1, seg_id=0, pcoord=np.array([[0], [0.25]], dtype=np.float32), weight=1.0 / 3.0),
            Segment(n_iter=1, seg_id=1, pcoord=np.array([[0], [0.75]], dtype=np.float32), weight=2.0 / 3.0),
        ]

        for _iround in range(nrounds):
            for segment in segments:
                segment.endpoint_type = Segment.SEG_ENDPOINT_UNSET

            self.we_driver.new_iteration()
            self.we_driver.assign(segments)
            self.we_driver.construct_next()

            assert len(self.we_driver.next_iter_binning[0]) == 1
            newseg = self.we_driver.next_iter_binning[0].pop()

            assert segments[newseg.parent_id].endpoint_type == Segment.SEG_ENDPOINT_CONTINUES
            assert segments[~newseg.parent_id].endpoint_type == Segment.SEG_ENDPOINT_MERGED

            selected_counts[newseg.parent_id] += 1

        print(selected_counts)
        assert (
            lb <= selected_counts[0] <= ub
        ), 'Incorrect proportion of histories selected.' 'this is expected about {:%} of the time; retry test.'.format(alpha)

    def test_split_with_adjust(self):
        # this is a split followed by merge
        self.system.bin_target_counts = np.array([5, 5])
        segments = [
            self.segment(1.5, 0.5, weight=0.125),
            self.segment(1.5, 0.5, weight=0.125),
            self.segment(0.0, 1.5, weight=0.375),
            self.segment(0.0, 1.5, weight=0.375),
        ]
        self.we_driver.new_iteration()
        self.we_driver.assign(segments)
        self.we_driver.construct_next()
        assert len(self.we_driver.next_iter_binning[0]) == 5
        assert len(self.we_driver.next_iter_binning[1]) == 5
        assert abs(sum(seg.weight for seg in self.we_driver.next_iter_binning[0]) - 0.25) < 5 * EPS
        assert abs(sum(seg.weight for seg in self.we_driver.next_iter_binning[1]) - 0.75) < 5 * EPS

        for ibin in range(2):
            for segment in self.we_driver.next_iter_binning[ibin]:
                print(segment)
                assert segment.n_iter == 2
                assert segment.parent_id is not None
                assert segment.status == Segment.SEG_STATUS_PREPARED

    def test_split_with_adjust_istates(self):
        # this is a split followed by merge, for segments which are initial states
        self.system.bin_target_counts = np.array([5, 5])
        # segments = [
        # self.segment(1.5, 0.5, weight=0.125),
        # self.segment(1.5, 0.5, weight=0.125),
        # self.segment(0.0, 1.5, weight=0.375),
        # self.segment(0.0, 1.5, weight=0.375),
        # ]
        self.we_driver.new_iteration()
        self.we_driver._prep_we()
        self.we_driver.used_initial_states[-1] = None
        self.we_driver.used_initial_states[-2] = None

        for ibin, bin in enumerate(self.we_driver.next_iter_binning):
            pc = np.array([[0.5 + ibin], [0.0]])
            for iseg in range(6):
                segment = Segment(n_iter=1, seg_id=None, weight=1.0 / 12.0, parent_id=-(ibin + 1), pcoord=pc)
                bin.add(segment)

        # for ibin in range(len(self.we_driver.next_iter_binning)):
        # This will raise KeyError if initial state tracking is done improperly
        #     self.we_driver._adjust_count(ibin)

        for ibin, bin in enumerate(self.we_driver.next_iter_binning):
            target_count = self.we_driver.bin_target_counts[ibin]
            subgroups = self.we_driver.subgroup_function(self.we_driver, ibin, **self.we_driver.subgroup_function_kwargs)
            self.we_driver._adjust_count(bin, subgroups, target_count)

        assert len(self.we_driver.next_iter_binning[0]) == 5
        assert len(self.we_driver.next_iter_binning[1]) == 5

    def test_recycle(self):
        segments = [self.segment(0.0, 1.5, weight=0.5), self.segment(0.0, 0.5, weight=0.5)]
        tstate = TargetState('recycle', [1.5], 0)
        istate = InitialState(0, 0, 0, pcoord=[0.0])

        self.we_driver.new_iteration(initial_states=[istate], target_states=[tstate])
        n_needed = self.we_driver.assign(segments)
        assert n_needed == 0
        self.we_driver.construct_next()

        n_recycled = len(list(self.we_driver.recycling_segments))
        assert n_recycled == 1

        assert len(self.we_driver.next_iter_binning[0]) == 4
        assert len(self.we_driver.next_iter_binning[1]) == 0
        assert abs(sum(seg.weight for seg in self.we_driver.next_iter_binning[0]) - 1.0) < 4 * EPS
        assert np.allclose([seg.weight for seg in self.we_driver.next_iter_binning[0]], [0.25 for _i in range(4)])
        assert segments[0].endpoint_type == Segment.SEG_ENDPOINT_RECYCLED

    def test_multiple_merge(self):
        # This weight and count combination is known to trigger a split to 51
        # followed by a count adjustment to 50 (thanks to Josh Adelman)
        segment = self.segment(0.0, 0.5, weight=0.9999999999970001)

        # Initial state ID 0
        segment.parent_id = -1
        segment.wtg_parent_ids = set([-1])
        assert segment.initpoint_type == segment.SEG_INITPOINT_NEWTRAJ

        self.system.bin_target_counts = np.array([50, 50])
        self.we_driver.new_iteration()
        self.we_driver.assign([segment])
        self.we_driver.construct_next()

        assert len(self.we_driver.next_iter_binning[0]) == 50

    def check_populate_initial(self, prob, target_counts):
        istate = InitialState(0, 0, 0, pcoord=[0.0])
        self.system.bin_target_counts = np.array([target_counts, target_counts])

        self.we_driver.populate_initial([istate], [prob], system=self.system)
        assert len(self.we_driver.next_iter_binning[0]) == target_counts

    @pytest.mark.skip
    def test_populate_initial(self):
        for prob in [0.1, 1.0 / 3.0, 0.9999999999970001, 1.0]:
            for tcount in range(30, 60):
                yield self.check_populate_initial, prob, tcount

    # TODO: add test for seeding the flux matrix based on recycling
    # TODO: add test for split after merge in adjust count



import numpy as np

import westpa
import westpa.core.yamlcfg as ycf
from westpa.core.binning import RectilinearBinMapper


class TESTSystem(ycf.YAMLSystem):
    def initialize(self):
        self.pcoord_ndim = 1
        self.pcoord_dtype = np.float32
        self.pcoord_len = 5
        self.bin_mapper = RectilinearBinMapper([list(np.arange(0.0, 10.1, 0.1))])
        self.bin_target_counts = np.empty((self.bin_mapper.nbins,), np.int_)
        self.bin_target_counts[...] = 10
        self.test_variable_2 = "And I'm the second one"


# YAML Front end tests

# Implemented basic tests
#  - returns the correct system
#    given a system driver
#  - returns the correct system
#    given a yaml system
#  - returns the correct system
#    given both


# A class to test both paths at the same time
# if it works we assure we can load the driver
# AND overwrite it properly
class TestYAMLFrontEnd:
    def testYAMLFEDriver(self):
        '''
        Test method to ensure the YAML system generator works as
        advertised
        '''

        # First the objects that will be used for testing
        rc = westpa.rc
        yamlConf = ycf.YAMLConfig()
        # A sample dictionary from a yaml file
        test_dict = {
            "west": {
                "system": {
                    "driver": "tests.test_yamlfe.TESTSystem",
                    "system_options": {
                        "pcoord_ndim": 2,
                        "test_variable": "I'm a test variable",
                        "pcoord_len": 10,
                        "pcoord_dtype": np.float32,
                        "bin_target_counts": 10,
                        "bins": {"type": "RectilinearBinMapper", "boundaries": [[0.0, 0.5, 1.5, 2.5, 3.5, 'inf']]},
                    },
                }
            }
        }
        yamlConf._data = test_dict
        rc.config = yamlConf

        self.system = rc.new_system_driver()

        system = self.system
        # Assert we have the right options
        # This needs some more documentation and alerts for the assertions
        assert system.pcoord_ndim == 2
        assert system.test_variable == "I'm a test variable"
        # This one in particular checks if the bins are passed correctly
        assert (system.bin_mapper.boundaries == np.array([[0.0, 0.5, 1.5, 2.5, 3.5, 'inf']], dtype=np.float32)).all()
        assert system.pcoord_len == 10
        assert system.pcoord_dtype == np.float32
        ## These should be the same as the original
        assert system.test_variable_2 == "And I'm the second one"

    def testYAMLFEConfig(self):
        # First the objects that will be used for testing
        rc = westpa.rc
        yamlConf = ycf.YAMLConfig()
        # A sample dictionary from a yaml file
        test_dict = {
            "west": {
                "system": {
                    "system_options": {
                        "pcoord_ndim": 2,
                        "test_variable": "I'm a test variable",
                        "pcoord_len": 10,
                        "pcoord_dtype": np.float32,
                        "bin_target_counts": 10,
                        "bins": {"type": "RectilinearBinMapper", "boundaries": ["np.arange(0.0, 5.0, 0.5)"]},
                    }
                }
            }
        }
        yamlConf._data = test_dict
        rc.config = yamlConf

        self.system = rc.new_system_driver()

        system = self.system
        # Assert we have the right options
        # This needs some more documentation and alerts for the assertions
        assert system.pcoord_ndim == 2
        assert system.test_variable == "I'm a test variable"
        # This one in particular checks if the bins are passed correctly
        assert (system.bin_mapper.boundaries == np.arange(0.0, 5.0, 0.5)).all()
        assert system.pcoord_len == 10
        assert system.pcoord_dtype == np.float32

        # Test __setitem__() method of YAMLConfig()
        rc.config['west', 'propagation', 'max_total_iteration'] = 1000

        assert rc.config['west', 'propagation', 'max_total_iteration'] == 1000



import numpy as np


from . import UncertMath


class ClusterList:
    def __init__(self, ratios, nbins):
        super().__init__()
        self.nbins = nbins
        self.ratios = ratios

        # Create an array to hold bin assignments and initial set all to -1 == not clustered
        self.bin_assign = np.empty((nbins,), dtype=int)
        self.bin_assign.fill(-1)

        # Initialize an ucert container to hold per bin information; initially mask all elements
        # note: masking is implicit since rates set to 0
        dum_data = np.zeros((nbins,))
        self.bin_data = UncertMath.UncertContainer(dum_data.copy(), dum_data.copy(), dum_data.copy())

        self.cluster_id = 0  # ID of newest cluster
        self.cluster_contents = {}  # Dictionary containing sets of bin ids with keys = cluster ids

    def join(self, pairs):
        """Join clusters given a tuple (i,j) of bin pairs"""

        for i, j in zip(*pairs):
            # Both bins not joined
            if self.bin_assign[i] == -1 and self.bin_assign[j] == -1:
                # Create new cluster
                self.bin_assign[i] = self.cluster_id
                self.bin_assign[j] = self.cluster_id

                self.cluster_contents[self.cluster_id] = {i, j}
                self.cluster_id += 1

                rij = self.ratios[i, j]
                denom = rij + 1.0
                self.bin_data[i] = rij / denom  # relative probability for bin i
                self.bin_data[j] = denom.recip()  # relative probability for bin j

            # Only one bin previously assigned to a cluster
            elif self.bin_assign[i] == -1 or self.bin_assign[j] == -1:
                if self.bin_assign[i] == -1:
                    idum, jdum = i, j
                else:
                    idum, jdum = j, i

                jclust = self.bin_assign[jdum]
                jclust_mid = np.where(self.bin_assign == jclust)[0]  # index of bins in jclust

                rik = self.ratios[idum, jclust_mid]
                pk = self.bin_data[jclust_mid]
                piTmp = rik * pk  # estimate for p_idum / P_cluster based on 'path' through bin k
                # Note that here P_cluster is value before addition of bin idum
                piTmp_avg = piTmp.weighted_average(axis=0)

                # now, compute relative prob of each bin in *new* cluster (including bin idum)
                denom = piTmp_avg + 1.0
                self.bin_data[idum] = piTmp_avg / denom

                # Update bins already in cluster
                self.bin_data[jclust_mid] = self.bin_data[jclust_mid] / denom

                # Move bin idum into cluster jclust
                self.bin_assign[idum] = jclust
                self.cluster_contents[jclust].update({idum})

            # Both bins previously assigned to different cluster; Join clusters
            elif not self.bin_assign[i] == self.bin_assign[j]:
                iclust = self.bin_assign[i]
                jclust = self.bin_assign[j]

                iclust_mid = np.where(self.bin_assign == iclust)[0]  # indx of bins in cluster i
                jclust_mid = np.where(self.bin_assign == jclust)[0]  # indx of bins in cluster j

                niclust = iclust_mid.size
                njclust = jclust_mid.size

                dum_data = np.zeros((niclust * njclust,))
                ij_cluster_ratio = UncertMath.UncertContainer(dum_data.copy(), dum_data.copy(), dum_data.copy())

                for count, im in enumerate(iclust_mid):
                    rij = self.ratios[im, jclust_mid]
                    pi = self.bin_data[im]
                    pj = self.bin_data[jclust_mid]

                    ij_cluster_ratio[count * njclust : (count + 1) * njclust] = rij * pj / pi

                ij_cluster_ratio = ij_cluster_ratio.weighted_average(axis=0)

                idenom = ij_cluster_ratio.recip() + 1.0
                jdenom = ij_cluster_ratio + 1.0

                self.bin_data[iclust_mid] = self.bin_data[iclust_mid] / idenom
                self.bin_data[jclust_mid] = self.bin_data[jclust_mid] / jdenom

                # Join all bins in cluster j into cluster iclust
                self.bin_assign[jclust_mid] = iclust

                # Move contents of jclust into iclust
                self.cluster_contents[iclust].update(self.cluster_contents[jclust])

                # Clear contents of jclust
                self.cluster_contents[jclust].clear()

                if len(self.cluster_contents[iclust]) == self.nbins:
                    break

    def join_simple(self, pairs):
        """Join clusters using direct ratios given a tuple (i,j) of bin pairs"""

        for i, j in zip(*pairs):
            # Both bins not joined
            if self.bin_assign[i] == -1 and self.bin_assign[j] == -1:
                # Create new cluster
                self.bin_assign[i] = self.cluster_id
                self.bin_assign[j] = self.cluster_id

                self.cluster_contents[self.cluster_id] = {i, j}
                self.cluster_id += 1

                rij = self.ratios[i, j]
                denom = rij + 1.0
                self.bin_data[i] = rij / denom  # relative probability for bin i
                self.bin_data[j] = denom.recip()  # relative probability for bin j

            # Only one bin previously assigned to a cluster
            elif self.bin_assign[i] == -1 or self.bin_assign[j] == -1:
                if self.bin_assign[i] == -1:
                    idum, jdum = i, j
                else:
                    idum, jdum = j, i

                jclust = self.bin_assign[jdum]
                rik = self.ratios[idum, jdum]
                pk = self.bin_data[jdum]
                piTmp = rik * pk  # estimate for p_idum / P_cluster based on 'path' through bin k
                # Note that here P_cluster is value before addition of bin idum
                # now, compute relative prob of each bin in *new* cluster (including bin idum)
                denom = piTmp + 1.0
                self.bin_data[idum] = piTmp / denom

                # Update bins already in cluster
                jclust_mid = np.where(self.bin_assign == jclust)  # index of bins in jclust
                self.bin_data[jclust_mid] = self.bin_data[jclust_mid] / denom

                # Move bin idum into cluster jclust
                self.bin_assign[idum] = jclust
                self.cluster_contents[jclust].update({idum})

            # Both bins previously assigned to different cluster; Join clusters
            elif not self.bin_assign[i] == self.bin_assign[j]:
                iclust = self.bin_assign[i]
                jclust = self.bin_assign[j]
                rij = self.ratios[i, j]
                pi = self.bin_data[i]
                pj = self.bin_data[j]
                ij_cluster_ratio = rij * pj / pi
                idenom = ij_cluster_ratio.recip() + 1.0
                jdenom = ij_cluster_ratio + 1.0

                iclust_mid = np.where(self.bin_assign == iclust)
                self.bin_data[iclust_mid] = self.bin_data[iclust_mid] / idenom

                jclust_mid = np.where(self.bin_assign == jclust)
                self.bin_data[jclust_mid] = self.bin_data[jclust_mid] / jdenom

                # Join all bins in cluster j into cluster iclust
                self.bin_assign[jclust_mid] = iclust

                # Move contents of jclust into iclust
                self.cluster_contents[iclust].update(self.cluster_contents[jclust])

                # Clear contents of jclust
                self.cluster_contents[jclust].clear()

                if len(self.cluster_contents[iclust]) == self.nbins:
                    break



from . import _assign
from . import assign, bins

from .assign import (
    NopMapper,
    FuncBinMapper,
    PiecewiseBinMapper,
    RectilinearBinMapper,
    RecursiveBinMapper,
    VectorizingFuncBinMapper,
    VoronoiBinMapper,
)

from .mab import map_mab, MABBinMapper
from .binless import map_binless, BinlessMapper

from .mab_driver import MABDriver
from .mab_manager import MABSimManager
from .binless_manager import BinlessSimManager
from .binless_driver import BinlessDriver

from ._assign import accumulate_labeled_populations, assign_and_label, accumulate_state_populations_from_labeled
from ._assign import assignments_list_to_table

from .assign import coord_dtype, index_dtype
from .bins import Bin


__all__ = [
    '_assign',
    'assign',
    'bins',
    'NopMapper',
    'FuncBinMapper',
    'PiecewiseBinMapper',
    'RectilinearBinMapper',
    'RecursiveBinMapper',
    'VectorizingFuncBinMapper',
    'VoronoiBinMapper',
    'map_mab',
    'map_binless',
    'MABBinMapper',
    'BinlessMapper',
    'MABDriver',
    'MABSimManager',
    'BinlessDriver',
    'BinlessSimManager',
    'accumulate_labeled_populations',
    'assign_and_label',
    'accumulate_state_populations_from_labeled',
    'assignments_list_to_table',
    'coord_dtype',
    'index_dtype',
    'Bin',
]



'''
Kinetics analysis library
'''

import logging

from .rate_averaging import RateAverager  # noqa

from . import _kinetics  # noqa
from ._kinetics import (  # noqa
    calculate_labeled_fluxes,
    labeled_flux_to_rate,
    calculate_labeled_fluxes_alllags,
    nested_to_flat_matrix,
    nested_to_flat_vector,
    flat_to_nested_matrix,
    flat_to_nested_vector,
    find_macrostate_transitions,
    sequence_macro_flux_to_rate,
)
from .events import WKinetics  # noqa


log = logging.getLogger(__name__)



import argparse
import os
from unittest import TestCase
from unittest.mock import MagicMock
import tempfile
import pytest

import numpy as np

import westpa
from westpa.core.binning.assign import RectilinearBinMapper
from westpa.core.segment import Segment
from westpa.core.states import BasisState
from westpa.core.sim_manager import PropagationError


def dummy_callback_one(self):
    pass


class TestSimManager(TestCase):
    def setUp(self):
        parser = argparse.ArgumentParser()
        westpa.rc.add_args(parser)

        here = os.path.dirname(__file__)
        os.environ['WEST_SIM_ROOT'] = os.path.join(here, 'fixtures', 'odld')

        config_file_name = os.path.join(here, 'fixtures', 'odld', 'west.cfg')
        args = parser.parse_args(['-r={}'.format(config_file_name)])
        westpa.rc.process_args(args)
        self.sim_manager = westpa.rc.get_sim_manager()
        self.test_dir = tempfile.mkdtemp()
        self.hdf5 = os.path.join(self.test_dir, "west.h5")
        self.basis_states = [BasisState(label="label", probability=1.0)]
        self.segments = [self.segment(0.0, 1.5, weight=0.125) for _i in range(4)] + [
            self.segment(1.5, 0.5, weight=0.125) for _i in range(4)
        ]
        self.sim_manager.we_driver.new_iteration()
        self.sim_manager.we_driver.assign(self.segments)
        self.sim_manager.we_driver.construct_next()
        self.sim_manager.segments = {segment.seg_id: segment for segment in self.segments}
        self.sim_manager.incomplete_segments = self.sim_manager.segments
        self.sim_manager.current_iter_istates = self.sim_manager.segments
        self.sim_manager.completed_segments = self.sim_manager.segments
        self.sim_manager.report_bin_statistics = MagicMock(return_value=True)

        data = self.sim_manager.we_driver.rc.get_data_manager()
        data.we_h5filename = self.hdf5
        data.prepare_backing()
        data.create_ibstate_group([])
        data.create_initial_states(1)
        data.save_target_states([])
        data.update_segments = MagicMock(return_value=None)

        n_iter = 0
        it_name = data.iter_group_name(n_iter)
        for group in ["seg_index", "parents", "ibstates", "pcoord"]:
            data.we_h5file.create_group(it_name + "/" + group)
        data.get_new_weight_data = MagicMock(return_value=None)
        data.get_segments = MagicMock(return_value=self.segments)
        self.sim_manager.we_driver.rc.get_data_manager = MagicMock(return_value=data)
        self.sim_manager.n_iter = n_iter

    def tearDown(self):
        westpa.rc._sim_manager = None
        westpa.rc._system = None
        westpa.rc._data_manager = None
        del os.environ['WEST_SIM_ROOT']

    def dummy_callback_one(self):
        system = self.sim_manager.system
        bounds = [0.0, 1.0, 2.0, 3.0]
        system.bin_mapper = RectilinearBinMapper([bounds])

    def dummy_callback_two(self):
        system = self.sim_manager.system
        bounds = [0.0, 1.0, 2.0, 5.0]
        system.bin_mapper = RectilinearBinMapper([bounds])

    def segment(self, init_pcoord, final_pcoord, weight=1.0):
        segment = Segment(n_iter=1, seg_id=1123, pcoord=self.sim_manager.system.new_pcoord_array(), weight=weight)
        segment.pcoord[0] = init_pcoord
        segment.pcoord[1] = final_pcoord
        segment.parent_id = 1
        return segment

    def test_sim_manager(self):
        self.assertEqual(self.sim_manager.n_propagated, 0)
        self.assertEqual(len(self.sim_manager._callback_table), 0)

    def test_register_callback(self):
        hook = self.sim_manager.prepare_new_iteration

        self.sim_manager.register_callback(hook, self.dummy_callback_one, 3)
        self.sim_manager.register_callback(hook, self.dummy_callback_two, 0)
        self.sim_manager.register_callback(
            hook, dummy_callback_one, 3
        )  # Same name and priority, but different function, should be added
        self.sim_manager.register_callback(
            hook, self.dummy_callback_one, 2
        )  # Duplicate should never be added, even with different priority
        self.assertTrue(hook in self.sim_manager._callback_table)

        callbacks = self.sim_manager._callback_table.get(hook, [])

        assert len(callbacks) == 3  # Make sure only 3 added.

        self.assertTrue((3, self.dummy_callback_one.__name__, self.dummy_callback_one) in callbacks)  # noqa
        self.assertTrue((0, self.dummy_callback_two.__name__, self.dummy_callback_two) in callbacks)  # noqa
        self.assertTrue((3, dummy_callback_one.__name__, dummy_callback_one) in callbacks)  # noqa

    def test_invoke_callback(self):
        hook = self.sim_manager.prepare_new_iteration

        self.sim_manager.register_callback(hook, self.dummy_callback_one, 3)
        self.sim_manager.register_callback(hook, self.dummy_callback_two, 0)

        self.sim_manager.invoke_callbacks(hook)

        system = self.sim_manager.system
        self.assertTrue(np.all(system.bin_mapper.boundaries == np.array([0.0, 1.0, 2.0, 3.0])))  # noqa

    def test_process_config(self):
        self.sim_manager.process_config()
        self.assertTrue(self.sim_manager.do_gen_istates)
        self.assertEqual(self.sim_manager.propagator_block_size, 10000)
        self.assertFalse(self.sim_manager.save_transition_matrices)
        self.assertEqual(self.sim_manager.max_run_walltime, 10800)
        self.assertEqual(self.sim_manager.max_total_iterations, 100)

    def test_load_plugins(self):
        self.sim_manager.load_plugins()

    def test_report_bin_statistics(self):
        self.sim_manager.report_bin_statistics([0.0, 1.0, 2.0, 5.0])

    def test_get_bstate_pcoords(self):
        self.sim_manager.get_bstate_pcoords(self.basis_states)

    def test_report_basis_states(self):
        self.sim_manager.report_basis_states(self.basis_states)

    def test_report_target_states(self):
        self.sim_manager.report_target_states(self.basis_states)

    @pytest.mark.skip(reason="Cannot currently test WESimManager.initialize_simulation()")
    def test_initialize_simulation(self):
        # TODO: determine how to test self.simulation_manager.initialize_simulation()
        pass

    def test_prepare_iteration(self):
        self.sim_manager.prepare_new_iteration()
        self.sim_manager.prepare_iteration()

    def test_finalize_iteration(self):
        self.sim_manager.finalize_iteration()

    def test_get_istate_futures(self):
        self.sim_manager.get_istate_futures()

    def test_propagate(self):
        westpa.core.states.pare_basis_initial_states = MagicMock(return_value=([], []))
        self.sim_manager.propagate

    def test_save_bin_data(self):
        self.sim_manager.save_bin_data()

    def test_check_propagation(self):
        self.assertRaises(PropagationError, self.sim_manager.check_propagation)

    def test_run_we(self):
        self.sim_manager.run_we()

    def test_run(self):
        self.sim_manager.run()

    def test_prepare_run(self):
        self.sim_manager.prepare_run()

    def test_finalize_run(self):
        self.sim_manager.finalize_run()

    def test_pre_propagation(self):
        self.sim_manager.pre_propagation()

    def test_post_propagation(self):
        self.sim_manager.post_propagation()

    def test_pre_we(self):
        self.sim_manager.pre_we()

    def test_post_we(self):
        self.sim_manager.post_we()


class TestMABSimManager(TestSimManager):
    def setUp(self):
        parser = argparse.ArgumentParser()
        westpa.rc.add_args(parser)

        here = os.path.dirname(__file__)
        os.environ['WEST_SIM_ROOT'] = os.path.join(here, 'fixtures', 'odld')

        config_file_name = os.path.join(here, 'fixtures', 'odld', 'west_mab.cfg')
        args = parser.parse_args(['-r={}'.format(config_file_name)])
        westpa.rc.process_args(args)
        self.sim_manager = westpa.rc.get_sim_manager()
        self.test_dir = tempfile.mkdtemp()
        self.hdf5 = os.path.join(self.test_dir, "west.h5")
        self.basis_states = [BasisState(label="label", probability=1.0)]
        self.segments = [self.segment(0.0, 1.5, weight=0.125) for _i in range(4)] + [
            self.segment(1.5, 0.5, weight=0.125) for _i in range(4)
        ]
        self.sim_manager.we_driver.new_iteration()
        self.sim_manager.we_driver.assign(self.segments)
        self.sim_manager.we_driver.construct_next()
        self.sim_manager.segments = {segment.seg_id: segment for segment in self.segments}
        self.sim_manager.incomplete_segments = self.sim_manager.segments
        self.sim_manager.current_iter_istates = self.sim_manager.segments
        self.sim_manager.completed_segments = self.sim_manager.segments
        self.sim_manager.report_bin_statistics = MagicMock(return_value=True)

        data = self.sim_manager.we_driver.rc.get_data_manager()
        data.we_h5filename = self.hdf5
        data.prepare_backing()
        data.create_ibstate_group([])
        data.create_initial_states(1)
        data.save_target_states([])
        data.update_segments = MagicMock(return_value=None)

        n_iter = 0
        it_name = data.iter_group_name(n_iter)
        for group in ["seg_index", "parents", "ibstates", "pcoord"]:
            data.we_h5file.create_group(it_name + "/" + group)
        data.get_new_weight_data = MagicMock(return_value=None)
        data.get_segments = MagicMock(return_value=self.segments)
        self.sim_manager.we_driver.rc.get_data_manager = MagicMock(return_value=data)
        self.sim_manager.n_iter = n_iter


class TestBinlessSimManager(TestSimManager):
    def setUp(self):
        parser = argparse.ArgumentParser()
        westpa.rc.add_args(parser)

        here = os.path.dirname(__file__)
        os.environ['WEST_SIM_ROOT'] = os.path.join(here, 'fixtures', 'odld')

        config_file_name = os.path.join(here, 'fixtures', 'odld', 'west_binless.cfg')
        args = parser.parse_args(['-r={}'.format(config_file_name)])
        westpa.rc.process_args(args)
        self.sim_manager = westpa.rc.get_sim_manager()
        self.test_dir = tempfile.mkdtemp()
        self.hdf5 = os.path.join(self.test_dir, "west.h5")
        self.basis_states = [BasisState(label="label", probability=1.0)]
        self.segments = [self.segment(0.0, 1.5, weight=0.125) for _i in range(4)] + [
            self.segment(1.5, 0.5, weight=0.125) for _i in range(4)
        ]
        self.sim_manager.we_driver.new_iteration()
        self.sim_manager.we_driver.assign(self.segments)
        self.sim_manager.we_driver.construct_next()
        self.sim_manager.segments = {segment.seg_id: segment for segment in self.segments}
        self.sim_manager.incomplete_segments = self.sim_manager.segments
        self.sim_manager.current_iter_istates = self.sim_manager.segments
        self.sim_manager.completed_segments = self.sim_manager.segments
        self.sim_manager.report_bin_statistics = MagicMock(return_value=True)

        data = self.sim_manager.we_driver.rc.get_data_manager()
        data.we_h5filename = self.hdf5
        data.prepare_backing()
        data.create_ibstate_group([])
        data.create_initial_states(1)
        data.save_target_states([])
        data.update_segments = MagicMock(return_value=None)

        n_iter = 0
        it_name = data.iter_group_name(n_iter)
        for group in ["seg_index", "parents", "ibstates", "pcoord"]:
            data.we_h5file.create_group(it_name + "/" + group)
        data.get_new_weight_data = MagicMock(return_value=None)
        data.get_segments = MagicMock(return_value=self.segments)
        self.sim_manager.we_driver.rc.get_data_manager = MagicMock(return_value=data)
        self.sim_manager.n_iter = n_iter

    @pytest.mark.skip('Not configured')
    def test_run(self):
        self.sim_manager.run()



# along with WESTPA.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function,division
import cython
import numpy
cimport numpy

# Coordinates are assumed to be 32-bit floats, and bin indices are
# defined to be 16-bit ints.

from numpy import uint16, float32
from numpy cimport uint16_t, float32_t

ctypedef numpy.float32_t _fptype

ctypedef numpy.uint8_t bool_t
ctypedef float32_t coord_t
ctypedef uint16_t index_t
ctypedef numpy.float64_t weight_t

bool_dtype = numpy.bool_
internal_bool_dtype = numpy.uint8
index_dtype = numpy.uint16
coord_dtype = numpy.float32

@cython.boundscheck(False)
@cython.wraparound(False)
cpdef rectilinear_assign(coord_t[:,:] coords,
                        numpy.ndarray[bool_t,ndim=1,cast=True] mask,
                        index_t[:] output,
                        boundaries,
                        index_t[:] boundlens):

    '''For bins delimited by sets boundaries on a rectilinear grid (``boundaries``),
    assign coordinates to bins, assuming C ordering of indices within the grid.
    ``boundlens`` is the number of boundaries in each dimension.

    '''
    cdef:
        int icoord, idim, ibound, boundlen
        int ndim
        int ncoords = coords.shape[0]
        index_t index, stridefac
        coord_t bound
        coord_t cval

        numpy.ndarray[coord_t, ndim=1] boundvec
        numpy.ndarray[numpy.uintp_t, ndim=1] boundvecs
        coord_t* bvec

    # We assume greater locality across boundary vectors than across the
    # coordinate array. To exploit this locality, and only have to
    # relinquish/reacquire the GIL once, we extract pointers to the first
    # element of each boundary vector; we then release the GIL and go to
    # town on the entire data set.

    ndim = len(boundaries)
    boundvecs = numpy.empty((ndim,), dtype=numpy.uintp)

    for 0 <= idim < ndim:
        boundvec = boundaries[idim]
        boundvecs[idim] = <numpy.uintp_t> &boundvec[0]

    with nogil:
        for icoord in range(ncoords):
            if not mask[icoord]:
                continue

            output[icoord] = 0
            stridefac = 1

            # backwards iteration needs signed values, so that the final != -1 works
            for idim in range(ndim-1,-1,-1):
                found = 0
                cval = coords[icoord,idim]
                boundlen = boundlens[idim]
                bvec = <coord_t*> boundvecs[idim]

                if cval < bvec[0] or cval >= bvec[boundlen-1]:
                    with gil:
                        raise ValueError('coordinate value {} is out of bin space in dimension {}'.format(cval,idim))

                for ibound in range(1,boundlen):
                    if cval < bvec[ibound]:
                        index = ibound-1
                        break

                output[icoord] += index * stridefac
                stridefac *= boundlen-1

@cython.boundscheck(False)
@cython.wraparound(False)
cpdef testfunc(coord_t[:,:] coords,
               numpy.ndarray[bool_t, ndim=1, cast=True] mask,
               index_t[:] output):
    cdef:
        index_t icoord

    for icoord in range(len(coords)):
        if mask[icoord]:
            if coords[icoord,0] < 0.5:
                output[icoord] = 0
            else:
                output[icoord] = 1

# optimized function applications
@cython.boundscheck(False)
@cython.wraparound(False)
cpdef apply_down(func,
                 args,
                 kwargs,
                 numpy.ndarray[coord_t, ndim=2] coords,
                 numpy.ndarray[bool_t, ndim=1, cast=True] mask,
                 index_t[:] output):
    '''Apply func(coord, *args, **kwargs) to each input coordinate tuple,
    skipping any for which mask is false and writing results to output.'''
    cdef:
        Py_ssize_t i, n

    n = len(output)
    for i from 0 <= i < n:
        if mask[i]:
            output[i] = func(coords[i], *args, **kwargs)

@cython.boundscheck(False)
@cython.wraparound(False)
cpdef apply_down_argmin_across(func,
                               args,
                               kwargs,
                               func_output_len,
                               numpy.ndarray[coord_t, ndim=2] coords,
                               numpy.ndarray[bool_t, ndim=1, cast=True] mask,
                               index_t[:] output):
    '''Apply func(coord, *args, **kwargs) to each input coordinate tuple,
    skipping any for which mask is false and writing results to output.'''
    cdef:
        Py_ssize_t icoord, iout, ncoord, nout,
        coord_t _min
        index_t _argmin
        numpy.ndarray[coord_t, ndim=1] func_output

    nout = func_output_len
    func_output = numpy.empty((func_output_len,), dtype=coord_dtype)

    ncoord = len(coords)
    for icoord from 0 <= icoord < ncoord:
        if mask[icoord]:
            func_output = func(coords[icoord], *args, **kwargs)
            if len(func_output) != func_output_len:
                raise TypeError('function returned a vector of length {} (expected length {})'
                                .format(len(func_output), func_output_len))

            # find minimum value
            _min = func_output[0]
            _argmin = 0
            for iout from 1 <= iout < nout:
                if func_output[iout] < _min:
                    _min = func_output[iout]
                    _argmin = iout

            output[icoord] = _argmin

# optimized lookup table routine
@cython.boundscheck(False)
@cython.wraparound(False)
cpdef output_map(index_t[:] output,
                 index_t[:] omap,
                 numpy.ndarray[bool_t, ndim=1, cast=True] mask):
    '''For each output for which mask is true, execute output[i] = omap[output[i]]'''

    cdef:
        Py_ssize_t i, ncoords, nmappings
        index_t o

    ncoords = len(output)
    nmappings = len(omap)
    with nogil:
        for i from 0 <= i < ncoords:
            if mask[i]:
                o = output[i]
                if o >= nmappings:
                    with gil:
                        raise IndexError('value {} not available in output table'.format(o))
                output[i] = omap[o]

@cython.boundscheck(False)
@cython.wraparound(False)
cpdef assign_and_label(Py_ssize_t nsegs_lb,
                       Py_ssize_t nsegs_ub,
                       long[:] parent_ids, # only for given segments
                       object assign,
                       Py_ssize_t nstates,
                       index_t[:] state_map,
                       index_t[:] last_labels, # must be for all segments
                       object pcoords, # only for given segments
                       bint subsample
                       ):
    '''Assign trajectories to bins and last-visted macrostates for each timepoint.'''

    cdef:
        Py_ssize_t ipt, nsegs, npts, iseg
        index_t[:,:] _assignments, _trajlabels, _statelabels
        index_t[:] seg_assignments
        long seg_id, parent_id, msegid
        index_t ptlabel

    nsegs = nsegs_ub - nsegs_lb
    npts = pcoords.shape[1]
    # We want to enable some subsampling, for steadystate simulations...
    # And we need to keep them alive through the merge.
    assignments = numpy.empty((nsegs,npts), index_dtype)
    trajlabels = numpy.empty((nsegs,npts), index_dtype)
    statelabels = numpy.empty((nsegs,npts), index_dtype)

    _assignments = assignments
    _trajlabels = trajlabels
    _statelabels = statelabels
    mask = numpy.ones((npts,), numpy.bool_)

    for iseg in range(nsegs):
        assign(pcoords[iseg,:], mask, assignments[iseg,:])

    if subsample is True:
        with nogil:
            for iseg in range(nsegs):
                for ipt in range(npts-2):
                    # We want to 'destroy' all assignment information that isn't the first or last point.
                    _assignments[iseg,ipt+1] = _assignments[iseg,0]

    if state_map is not None:
        if subsample is False:
            with nogil:
                for iseg in range(nsegs):
                    seg_id = iseg+nsegs_lb
                    parent_id = parent_ids[iseg]
                    for ipt in range(npts):
                        ptlabel = state_map[_assignments[iseg,ipt]]
                        _statelabels[iseg,ipt] = ptlabel
                        if ptlabel == nstates: # unknown state/transition region
                            if ipt == 0:
                                if parent_id < 0:
                                    # We have started a trajectory in a transition region
                                    _trajlabels[iseg,ipt] = nstates
                                    _statelabels[iseg,ipt] = nstates
                                else:
                                    # We can inherit the ending point from the previous iteration
                                    # This should be nstates (unknown_state) for the first iteration
                                    _trajlabels[iseg,ipt] = last_labels[parent_id]
                            else:
                                # We are currently in a transition region, but we care about the last state we visited,
                                # so inherit that state from the previous point
                                _trajlabels[iseg,ipt] = _trajlabels[iseg,ipt-1]
                        else:
                            _trajlabels[iseg,ipt] = ptlabel
        else:
            with nogil:
                for iseg in range(nsegs):
                    seg_id = iseg+nsegs_lb
                    parent_id = parent_ids[iseg]
                    for ipt in range(npts):
                        if ipt == 0:
                            ptlabel = state_map[_assignments[iseg,ipt]]
                            _statelabels[iseg,ipt] = ptlabel
                            if ptlabel == nstates: # unknown state/transition region
                                if parent_id < 0:
                                    # We have started a trajectory in a transition region
                                    _trajlabels[iseg,ipt] = nstates
                                    _statelabels[iseg,ipt] = nstates
                                else:
                                    # We can inherit the ending point from the previous iteration
                                    # This should be nstates (unknown_state) for the first iteration
                                    _trajlabels[iseg,ipt] = last_labels[parent_id]
                            else:
                                _trajlabels[iseg,ipt] = ptlabel
                        if ipt != 0:
                            # We're no longer at the first point, so just nab the previous assignment.
                            _trajlabels[iseg,ipt] = _trajlabels[iseg,ipt-1]
                            _statelabels[iseg,ipt] = _statelabels[iseg,ipt-1]
                            # If we're at the last time point and we've had a transition, reflect it.
                            if ipt == npts - 1:
                                ptlabel = state_map[_assignments[iseg,ipt]]
                                if ptlabel != nstates:
                                    _statelabels[iseg,npts-1] = ptlabel
                                    _trajlabels[iseg,npts-1] = ptlabel

    else:
        trajlabels.fill(nstates)
        statelabels.fill(nstates)

    return assignments, trajlabels, statelabels

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
cpdef accumulate_labeled_populations(weight_t[:]  weights,
                                     index_t[:,:] bin_assignments,
                                     index_t[:,:] label_assignments,
                                     weight_t[:,:] labeled_bin_pops):
    '''For a set of segments in one iteration, calculate the average population in each bin, with
    separation by last-visited macrostate.'''
    cdef:
        Py_ssize_t nsegs, npts, nbins, nstates, seg_id, ipt
        index_t assignment, traj_assignment
        weight_t ptwt

    nstates = labeled_bin_pops.shape[0] # this will generally be n_valid_states + 1 (for unknown state)
    nbins = labeled_bin_pops.shape[1]   # this may be n_valid_bins + 1 (for unknown bin)
    nsegs = bin_assignments.shape[0]
    npts = bin_assignments.shape[1]

    with nogil:
        for seg_id in range(nsegs):
            ptwt = weights[seg_id] / npts
            for ipt in range(npts):
                assignment = bin_assignments[seg_id,ipt]
                if assignment >= nbins:
                    with gil:
                        raise ValueError('invalid bin assignment for segment {} point {}'.format(seg_id, ipt))

                traj_assignment = label_assignments[seg_id,ipt]
                if traj_assignment >= nstates:
                    with gil:
                        raise ValueError('invalid trajectory label for segment {} point {}'.format(seg_id, ipt))

                labeled_bin_pops[traj_assignment,assignment] += ptwt

@cython.boundscheck(False)
@cython.wraparound(False)
cpdef accumulate_state_populations_from_labeled(weight_t[:,:] labeled_bin_pops,
                                                index_t[:] state_map,
                                                weight_t[:] state_pops,
                                                check_state_map = True):
    cdef:
        Py_ssize_t nbins, nstates, ibin, ilabel, istate

    if state_pops.shape[0] != labeled_bin_pops.shape[0]:
        raise TypeError('shape mismatch')

    nstates = labeled_bin_pops.shape[0]
    nbins = labeled_bin_pops.shape[1]

    if check_state_map:
        for ibin in range(nbins):
            if state_map[ibin] > nstates:
                raise ValueError('invalid value in state map')

    with nogil:
        for ilabel in range(nstates):
            for ibin in range(nbins):
                istate = state_map[ibin]
                if istate >= nstates:
                    with gil:
                        raise ValueError('invalid state label {}'.format(istate))
                state_pops[state_map[ibin]] += labeled_bin_pops[ilabel,ibin]

@cython.wraparound(False)
cpdef assignments_list_to_table(Py_ssize_t nsegs, Py_ssize_t nbins, index_t[:] assignments):
    '''Convert a list of bin assignments (integers) to a boolean table indicating indicating
    if a given segment is in a given bin'''

    cdef:
        Py_ssize_t iseg, ibin
        bool_t[:,:] _output

    output = numpy.zeros((nsegs,nbins), internal_bool_dtype)
    _output = output

    for iseg in range(nsegs):
        _output[iseg,assignments[iseg]] = 1

    return output.astype(numpy.bool_, copy=False)



import logging
import numpy as np
import westpa
from westpa.core.binning import FuncBinMapper
from os.path import expandvars


log = logging.getLogger(__name__)


class MABBinMapper(FuncBinMapper):
    """
    Adaptively place bins in between minimum and maximum segments along
    the progress coordinte. Extrema and bottleneck segments are assigned
    to their own bins.

    """

    def __init__(
        self,
        nbins,
        direction=None,
        skip=None,
        bottleneck=True,
        pca=False,
        mab_log=False,
        bin_log=False,
        bin_log_path="$WEST_SIM_ROOT/binbounds.log",
    ):
        """
        Parameters
        ----------
        nbins : list of int
            List of int for nbins in each dimension.
        direction : Union(list of int, None), default: None
            List of int for 'direction' in each dimension.
            Direction options are as follows:
                0   : default split at leading and lagging boundaries
                1   : split at leading boundary only
                -1  : split at lagging boundary only
                86  : no splitting at either leading or lagging boundary
        skip : Union(list of int, None), default: None
            List of int for each dimension. Default None for skip=0.
            Set to 1 to 'skip' running mab in a dimension.
        bottleneck : bool, default: True
            Whether to turn on or off bottleneck walker splitting.
        pca : bool, default: False
            Can be True or False (default) to run PCA on pcoords before bin assignment.
        mab_log : bool, default: False
            Whether to output mab info to west.log.
        bin_log : bool, default: False
            Whether to output mab bin boundaries to bin_log_path file.
        bin_log_path : str, default: "$WEST_SIM_ROOT/binbounds.log"
            Path to output bin boundaries.

        """
        # Verifying parameters
        if nbins is None:
            raise ValueError("nbins_per_dim is missing")
        ndim = len(nbins)

        if direction is None:
            direction = [0] * ndim
        elif len(direction) != ndim:
            direction = [0] * ndim
            log.warning("Direction list is not the correct dimensions, setting to defaults.")

        if skip is None:
            skip = [0] * ndim
        elif len(skip) != ndim:
            skip = [0] * ndim
            log.warning("Skip list is not the correct dimensions, setting to defaults.")

        kwargs = dict(
            nbins_per_dim=nbins,
            direction=direction,
            skip=skip,
            bottleneck=bottleneck,
            pca=pca,
            mab_log=mab_log,
            bin_log=bin_log,
            bin_log_path=bin_log_path,
        )

        n_total_bins = self.determine_total_bins(**kwargs)

        super().__init__(map_mab, n_total_bins, kwargs=kwargs)

    def determine_total_bins(self, nbins_per_dim, direction, skip, bottleneck, **kwargs):
        """
        The following is neccessary because functional bin mappers need to "reserve"
        bins and tell the sim manager how many bins they will need to use, this is
        determined by taking all direction/skipping info into account.

        Parameters
        ----------
        nbins_per_dim : int
            Number of total bins in each direction.
        direction : list of int
            Direction in each dimension. See __init__ for more information.
        skip : list of int
            List of 0s and 1s indicating whether to skip each dimension.
        bottleneck : bool
            Whether to include separate bin for bottleneck walker(s).
        **kwargs : dict
            Arbitary keyword arguments. Contains unneeded MAB parameters.

        Returns
        -------
        n_total_bins : int
            Number of total bins.

        """
        n_total_bins = np.prod(nbins_per_dim)
        ndim = len(nbins_per_dim)
        for i in range(ndim):
            if skip[i] == 0:
                if direction[i] != 0:
                    n_total_bins += 1 + 1 * bottleneck
                else:
                    n_total_bins += 2 + 2 * bottleneck
            else:
                n_total_bins -= nbins_per_dim[i] - 1
                n_total_bins += 1 * ndim  # or else it will be one bin short
        return n_total_bins


def map_mab(coords, mask, output, *args, **kwargs):
    """
    Binning which adaptively places bins based on the positions of extrema segments and
    bottleneck segments, which are where the difference in probability is the greatest
    along the progress coordinate. Operates per dimension and places a fixed number of
    evenly spaced bins between the segments with the min and max pcoord values. Extrema and
    bottleneck segments are assigned their own bins.

    Parameters
    ----------
    coords : ndarray
        An array with pcoord and weight info.
    mask : ndarray
        Array of 1 (True) and 0 (False), to filter out unwanted segment info.
    output : list
        The main list that, for each segment, holds the bin assignment.
    *args : list
        Variable length arguments.
    **kwargs : dict
        Arbitary keyword arguments. Contains most of the MAB-needed parameters.

    Returns
    ------
    output : list
        The main list that, for each segment, holds the bin assignment.

    """

    # Argument Processing
    nbins_per_dim = kwargs.get("nbins_per_dim")
    ndim = len(nbins_per_dim)
    pca = kwargs.get("pca", False)
    bottleneck = kwargs.get("bottleneck", True)
    direction = kwargs.get("direction", ([0] * ndim))
    skip = kwargs.get("skip", ([0] * ndim))
    mab_log = kwargs.get("mab_log", False)
    bin_log = kwargs.get("bin_log", False)
    bin_log_path = kwargs.get("bin_log_path", "$WEST_SIM_ROOT/binbounds.log")

    if not np.any(mask):
        return output

    if skip is None:
        skip = [0] * ndim

    allcoords = np.copy(coords)
    allmask = np.copy(mask)

    weights = None
    isfinal = None
    splitting = False
    report = False

    # the segments should be sent in by the driver as half initial segments and half final segments
    # allcoords contains all segments
    # coords should contain ONLY final segments
    if coords.shape[1] > ndim:
        if coords[0, -1] == 0:
            report = True
        if coords.shape[1] > ndim + 1:
            isfinal = allcoords[:, ndim + 1].astype(np.bool_)
        else:
            isfinal = np.ones(coords.shape[0], dtype=np.bool_)
        coords = coords[isfinal, :ndim]
        weights = allcoords[isfinal, ndim + 0]
        mask = mask[isfinal]
        splitting = True

    if not np.any(mask):
        coords = allcoords[:, :ndim]
        mask = allmask
        weights = None
        splitting = False

    varcoords = np.copy(coords)
    originalcoords = np.copy(coords)
    if pca and len(output) > 1:
        colavg = np.mean(coords, axis=0)
        for i in range(len(coords)):
            for j in range(len(coords[i])):
                varcoords[i][j] = coords[i][j] - colavg[j]
        covcoords = np.cov(np.transpose(varcoords), aweights=weights)
        eigval, eigvec = np.linalg.eigh(covcoords)
        eigvec = eigvec[:, np.argmax(np.absolute(eigvec), axis=1)]
        for i in range(len(eigvec)):
            if eigvec[i, i] < 0:
                eigvec[:, i] = -1 * eigvec[:, i]
        for i in range(ndim):
            for j in range(len(output)):
                coords[j][i] = np.dot(varcoords[j], eigvec[:, i])

    maxlist = []
    minlist = []
    difflist = []
    flipdifflist = []
    for n in range(ndim):
        # identify the boundary segments
        maxcoord = np.max(coords[mask, n])
        mincoord = np.min(coords[mask, n])
        maxlist.append(maxcoord)
        minlist.append(mincoord)

        # detect the bottleneck segments, this uses the weights
        if splitting:
            temp = np.column_stack((originalcoords[mask, n], weights[mask]))
            sorted_indices = temp[:, 0].argsort()
            temp = temp[sorted_indices]
            for p in range(len(temp)):
                if temp[p][1] == 0:
                    temp[p][1] = 10**-323
            fliptemp = np.flipud(temp)

            difflist.append(None)
            flipdifflist.append(None)
            maxdiff = 0
            flipmaxdiff = 0
            for i in range(1, len(temp) - 1):
                comprob = 0
                flipcomprob = 0
                j = i + 1
                while j < len(temp):
                    comprob = comprob + temp[j][1]
                    flipcomprob = flipcomprob + fliptemp[j][1]
                    j = j + 1
                diff = -np.log(comprob) + np.log(temp[i][1])
                if diff > maxdiff:
                    difflist[n] = temp[i][0]
                    maxdiff = diff
                flipdiff = -np.log(flipcomprob) + np.log(fliptemp[i][1])
                if flipdiff > flipmaxdiff:
                    flipdifflist[n] = fliptemp[i][0]
                    flipmaxdiff = flipdiff

    if mab_log and report:
        westpa.rc.pstatus("################ MAB stats ################")
        westpa.rc.pstatus("minima in each dimension:      {}".format(minlist))
        westpa.rc.pstatus("maxima in each dimension:      {}".format(maxlist))
        westpa.rc.pstatus("direction in each dimension:   {}".format(direction))
        westpa.rc.pstatus("skip in each dimension:        {}".format(skip))
        westpa.rc.pstatus("###########################################")
        westpa.rc.pflush()

    # assign segments to bins
    # the total number of linear bins is the boundary base
    boundary_base = np.prod(nbins_per_dim)

    # the bottleneck base is offset by the number of boundary walkers,
    # which is two per dimension unless there is a direction specified
    # in a particluar dimension, then it's just one
    bottleneck_base = boundary_base
    n_bottleneck_filled = 0

    for i in range(0, ndim):
        # for single direction, 1 boundary walker
        if direction[i] == 1 or direction[i] == -1:
            bottleneck_base += 1
        # 2 boundary walkers with 0 direction
        elif direction[i] == 0:
            bottleneck_base += 2
        # for 86 direction, no boundary walkers so offset of 0
        elif direction[i] == 86:
            bottleneck_base += 0

    # if a dimension is being "skipped", leave only one bin total as
    # the offset
    for i in range(0, ndim):
        if skip[i] != 0:
            boundary_base -= nbins_per_dim[i] - 1

    for i in range(len(output)):
        if not allmask[i]:
            continue

        # special means either a boundary or bottleneck walker (not a walker in the linear space)
        special = False
        # this holder is the bin number, which only needs to be unique for different walker groups
        holder = 0
        if splitting:
            for n in range(ndim):
                coord = allcoords[i][n]

                # if skipped, just assign the walkers to the same bin (offset of boundary base)
                if skip[n] != 0:
                    holder = boundary_base + n
                    break

                # assign bottlenecks, taking directionality into account
                if bottleneck:
                    if direction[n] == -1:
                        if coord == flipdifflist[n]:
                            holder = bottleneck_base + n
                            special = True
                            n_bottleneck_filled += 1
                            break

                    if direction[n] == 1:
                        if coord == difflist[n]:
                            holder = bottleneck_base + n
                            special = True
                            n_bottleneck_filled += 1
                            break

                    # both directions when using 0 or with
                    # special value of 86 for no lead/lag split
                    if direction[n] == 0 or direction[n] == 86:
                        if coord == difflist[n]:
                            holder = bottleneck_base + n
                            special = True
                            n_bottleneck_filled += 1
                            break
                        elif coord == flipdifflist[n]:
                            holder = bottleneck_base + n + 1
                            special = True
                            n_bottleneck_filled += 1
                            break

                # assign boundary walkers, taking directionality into account
                if direction[n] == -1:
                    if coord == minlist[n]:
                        holder = boundary_base + n
                        special = True
                        break

                elif direction[n] == 1:
                    if coord == maxlist[n]:
                        holder = boundary_base + n
                        special = True
                        break

                elif direction[n] == 0:
                    if coord == minlist[n]:
                        holder = boundary_base + n
                        special = True
                        break
                    elif coord == maxlist[n]:
                        holder = boundary_base + n + 1
                        special = True
                        break

                # special value for direction with no lead/lag split
                elif direction[n] == 86:
                    # westpa.rc.pstatus(f"No lead/lag split for dim {n}")
                    # westpa.rc.pflush()
                    # nornmally adds to special bin but here just leaving it forever empty
                    # holder = boundary_base + n
                    break

        # the following are for the "linear" portion
        if not special:
            for n in range(ndim):
                # if skipped, it's added to the same bin as the special walkers above
                if skip[n] != 0:
                    holder = boundary_base + n
                    break

                coord = allcoords[i][n]
                nbins = nbins_per_dim[n]
                minp = minlist[n]
                maxp = maxlist[n]

                bins = np.linspace(minp, maxp, nbins + 1)
                bin_number = np.digitize(coord, bins) - 1

                if isfinal is None or not isfinal[i]:
                    if bin_number >= nbins:
                        bin_number = nbins - 1
                    elif bin_number < 0:
                        bin_number = 0
                elif bin_number >= nbins or bin_number < 0:
                    if np.isclose(bins[-1], coord):
                        bin_number = nbins - 1
                    elif np.isclose(bins[0], coord):
                        bin_number = 0
                    else:
                        raise ValueError("Walker out of boundary")

                holder += bin_number * np.prod(nbins_per_dim[:n])

        # output is the main list that, for each segment, holds the bin assignment
        output[i] = holder

    if bin_log and report:
        if westpa.rc.sim_manager.n_iter:
            with open(expandvars(bin_log_path), 'a') as bb_file:
                # Iteration Number
                bb_file.write(f'iteration: {westpa.rc.sim_manager.n_iter}\n')
                bb_file.write('bin boundaries: ')
                for n in range(ndim):
                    # Write binbounds per dim
                    bb_file.write(f'{np.linspace(minlist[n], maxlist[n], nbins_per_dim[n] + 1)}\t')
                    # Min/Max pcoord
                bb_file.write(f'\nmin/max pcoord: {minlist} {maxlist}\n')
                bb_file.write(f'bottleneck bins: {n_bottleneck_filled}\n')
                if n_bottleneck_filled > 0:
                    # Bottlenecks bins exist (passes any of the if bottleneck: checks)
                    bb_file.write(f'bottleneck pcoord: {flipdifflist} {difflist}\n\n')
                else:
                    bb_file.write('\n')

    return output



import numpy as np

from westpa.tools import WESTDataReader, IterRangeSelection, WESTSubcommand, ProgressIndicatorComponent

from westpa import mclib
from westpa.core import h5io

from westpa.tools.dtypes import iter_block_ci_dtype as ci_dtype


# A function to just help with creating future objects for the work manager.


def generate_future(work_manager, name, eval_block, kwargs):
    submit_kwargs = {'name': name}
    submit_kwargs.update(kwargs)
    future = work_manager.submit(eval_block, kwargs=submit_kwargs)
    return future


class WESTKineticsBase(WESTSubcommand):
    '''
    Common argument processing for w_direct/w_reweight subcommands.
    Mostly limited to handling input and output from w_assign.
    '''

    def __init__(self, parent):
        super().__init__(parent)

        self.data_reader = WESTDataReader()
        self.iter_range = IterRangeSelection()
        self.progress = ProgressIndicatorComponent()

        self.output_filename = None
        # This is actually applicable to both.
        self.assignment_filename = None

        self.output_file = None
        self.assignments_file = None

        self.evolution_mode = None

        self.mcbs_alpha = None
        self.mcbs_acalpha = None
        self.mcbs_nsets = None

        # Now we're adding in things that come from the old w_kinetics
        self.do_compression = True

    def add_args(self, parser):
        self.progress.add_args(parser)
        self.data_reader.add_args(parser)
        self.iter_range.include_args['iter_step'] = True
        self.iter_range.add_args(parser)

        iogroup = parser.add_argument_group('input/output options')
        iogroup.add_argument(
            '-a',
            '--assignments',
            default='assign.h5',
            help='''Bin assignments and macrostate definitions are in ASSIGNMENTS
                            (default: %(default)s).''',
        )

        iogroup.add_argument(
            '-o',
            '--output',
            dest='output',
            default=self.default_output_file,
            help='''Store results in OUTPUT (default: %(default)s).''',
        )

    def process_args(self, args):
        self.progress.process_args(args)
        self.data_reader.process_args(args)
        with self.data_reader:
            self.iter_range.process_args(args, default_iter_step=None)
        if self.iter_range.iter_step is None:
            # use about 10 blocks by default
            self.iter_range.iter_step = max(1, (self.iter_range.iter_stop - self.iter_range.iter_start) // 10)

        self.output_filename = args.output
        self.assignments_filename = args.assignments


# This provides some convenience functions, modified from w_kinavg, to help with calculating evolution and averages for observables with the mclib library in a consistent manner.
# It's used in both w_direct and w_reweight.
class AverageCommands(WESTKineticsBase):
    default_output_file = 'direct.h5'

    def __init__(self, parent):
        # Ideally, this is stuff general to all the calculations we want to perform.
        super().__init__(parent)
        self.kinetics_filename = None
        self.kinetics_file = None

    def add_args(self, parser):
        iogroup = parser.add_argument_group('input/output options')
        # self.default_kinetics_file will be picked up as a class attribute from the appropriate subclass
        # We can do this with the output file, too...
        # ... by default, however, we're going to use {direct/reweight}.h5 for everything.
        # Modules which are called with different default values will, of course, still use those.
        iogroup.add_argument(
            '-k',
            '--kinetics',
            default=self.default_kinetics_file,
            help='''Populations and transition rates are stored in KINETICS
                            (default: %(default)s).''',
        )

        cgroup = parser.add_argument_group('confidence interval calculation options')
        cgroup.add_argument(
            '--disable-bootstrap',
            '-db',
            dest='bootstrap',
            action='store_const',
            const=False,
            help='''Enable the use of Monte Carlo Block Bootstrapping.''',
        )
        cgroup.add_argument(
            '--disable-correl',
            '-dc',
            dest='correl',
            action='store_const',
            const=False,
            help='''Disable the correlation analysis.''',
        )
        cgroup.add_argument(
            '--alpha',
            type=float,
            default=0.05,
            help='''Calculate a (1-ALPHA) confidence interval'
                             (default: %(default)s)''',
        )
        cgroup.add_argument(
            '--autocorrel-alpha',
            type=float,
            dest='acalpha',
            metavar='ACALPHA',
            help='''Evaluate autocorrelation to (1-ACALPHA) significance.
                             Note that too small an ACALPHA will result in failure to detect autocorrelation
                             in a noisy flux signal. (Default: same as ALPHA.)''',
        )
        cgroup.add_argument('--nsets', type=int, help='''Use NSETS samples for bootstrapping (default: chosen based on ALPHA)''')

        cogroup = parser.add_argument_group('calculation options')
        cogroup.add_argument(
            '-e',
            '--evolution-mode',
            choices=['cumulative', 'blocked', 'none'],
            default='none',
            help='''How to calculate time evolution of rate estimates.
                             ``cumulative`` evaluates rates over windows starting with --start-iter and getting progressively
                             wider to --stop-iter by steps of --step-iter.
                             ``blocked`` evaluates rates over windows of width --step-iter, the first of which begins at
                             --start-iter.
                             ``none`` (the default) disables calculation of the time evolution of rate estimates.''',
        )
        cogroup.add_argument(
            '--window-frac',
            type=float,
            default=1.0,
            help='''Fraction of iterations to use in each window when running in ``cumulative`` mode.
                             The (1 - frac) fraction of iterations will be discarded from the start of each window.''',
        )

        mgroup = parser.add_argument_group('misc options')
        mgroup.add_argument(
            '--disable-averages',
            '-da',
            dest='display_averages',
            action='store_false',
            help='''Whether or not the averages should be printed to the console (set to FALSE if flag is used).''',
        )

    def process_args(self, args):
        self.kinetics_filename = args.kinetics

        # Disable the bootstrap or the correlation analysis.
        self.mcbs_enable = args.bootstrap if args.bootstrap is not None else True
        self.do_correl = args.correl if args.correl is not None else True
        self.mcbs_alpha = args.alpha
        self.mcbs_acalpha = args.acalpha if args.acalpha else self.mcbs_alpha
        self.mcbs_nsets = args.nsets if args.nsets else mclib.get_bssize(self.mcbs_alpha)

        self.display_averages = args.display_averages

        self.evolution_mode = args.evolution_mode
        self.evol_window_frac = args.window_frac
        if self.evol_window_frac <= 0 or self.evol_window_frac > 1:
            raise ValueError('Parameter error -- fractional window defined by --window-frac must be in (0,1]')

    def stamp_mcbs_info(self, dataset):
        dataset.attrs['mcbs_alpha'] = self.mcbs_alpha
        dataset.attrs['mcbs_acalpha'] = self.mcbs_acalpha
        dataset.attrs['mcbs_nsets'] = self.mcbs_nsets

    def open_files(self):
        self.output_file = h5io.WESTPAH5File(self.output_filename, 'a', creating_program=True)
        h5io.stamp_creator_data(self.output_file)
        self.assignments_file = h5io.WESTPAH5File(self.assignments_filename, 'r')  # , driver='core', backing_store=False)
        self.kinetics_file = h5io.WESTPAH5File(self.kinetics_filename, 'r')  # , driver='core', backing_store=False)
        if not self.iter_range.check_data_iter_range_least(self.assignments_file):
            raise ValueError('assignments data do not span the requested iterations')

    def open_assignments(self):
        # Actually, I should rename this, as we're not OPENING assignments.
        # This seems to be stuff we're going to be using a lot, so.
        self.nstates = self.assignments_file.attrs['nstates']
        self.nbins = self.assignments_file.attrs['nbins']
        self.state_labels = self.assignments_file['state_labels'][...]
        assert self.nstates == len(self.state_labels)
        self.start_iter, self.stop_iter, self.step_iter = (
            self.iter_range.iter_start,
            self.iter_range.iter_stop,
            self.iter_range.iter_step,
        )

        # Import for the reweighting code.

        # state_map = self.assignments_file['state_map'][...]

        # We've moved this into a different step so that it's compatible with
        # loading up from the all command.
        # Otherwise, we try to load the kinetics (since we're just mixing subclasses)
        # before it's actually run, and so we fail out.
        if not self.iter_range.check_data_iter_range_least(self.kinetics_file):
            raise ValueError('kinetics data do not span the requested iterations')

    def print_averages(self, dataset, header, dim=1):
        print(header)
        maxlabellen = max(list(map(len, self.state_labels)))
        for istate in range(self.nstates):
            if dim == 1:
                print(
                    '{:{maxlabellen}s}: mean={:21.15e} CI=({:21.15e}, {:21.15e}) * tau^-1'.format(
                        self.state_labels[istate],
                        dataset['expected'][istate],
                        dataset['ci_lbound'][istate],
                        dataset['ci_ubound'][istate],
                        maxlabellen=maxlabellen,
                    )
                )

            else:
                for jstate in range(self.nstates):
                    if istate == jstate:
                        continue
                    print(
                        '{:{maxlabellen}s} -> {:{maxlabellen}s}: mean={:21.15e} CI=({:21.15e}, {:21.15e}) * tau^-1'.format(
                            self.state_labels[istate],
                            self.state_labels[jstate],
                            dataset['expected'][istate, jstate],
                            dataset['ci_lbound'][istate, jstate],
                            dataset['ci_ubound'][istate, jstate],
                            maxlabellen=maxlabellen,
                        )
                    )

    def run_calculation(
        self, pi, nstates, start_iter, stop_iter, step_iter, dataset, eval_block, name, dim, do_averages=False, **extra
    ):
        # We want to use the same codepath to run a quick average as we do the longer evolution sets, so...
        if do_averages:
            start_pts = [start_iter, stop_iter]
        else:
            start_pts = list(range(start_iter, stop_iter, step_iter))
        # Our evolution dataset!
        if dim == 2:
            evolution_dataset = np.zeros((len(start_pts), nstates, nstates), dtype=ci_dtype)
        elif dim == 1:
            evolution_dataset = np.zeros((len(start_pts), nstates), dtype=ci_dtype)
        else:
            # Temp.
            print("What's wrong?")

        # This is appropriate for bootstrapped quantities, I think.

        if True:
            futures = []
            for iblock, start in enumerate(start_pts):
                stop = min(start + step_iter, stop_iter)
                if self.evolution_mode == 'cumulative' or do_averages is True:
                    windowsize = int(self.evol_window_frac * (stop - start_iter))
                    block_start = max(start_iter, stop - windowsize)
                else:  # self.evolution_mode == 'blocked'
                    block_start = start

                # Create a basic set of kwargs for this iteration slice.
                future_kwargs = dict(
                    iblock=iblock,
                    start=block_start,
                    stop=stop,
                    nstates=nstates,
                    mcbs_alpha=self.mcbs_alpha,
                    mcbs_nsets=self.mcbs_nsets,
                    mcbs_acalpha=self.mcbs_acalpha,
                    do_correl=self.do_correl,
                    name=name,
                    mcbs_enable=self.mcbs_enable,
                    data_input={},
                    **extra
                )

                # Slice up the datasets for this iteration slice.
                # We're assuming they're all h5io iter blocked datasets; it's up to the calling routine
                # to ensure this is true.

                # Actually, I'm less sure how to handle this for pre-calculated datasets.  Need to consider this.  But for now...
                for key, value in dataset.items():
                    try:
                        future_kwargs['data_input'][key] = (
                            value.iter_slice(block_start, stop) if hasattr(value, 'iter_slice') else value[block_start:stop]
                        )
                    except Exception:
                        future_kwargs['data_input'][key] = (
                            value.iter_slice(block_start, stop) if hasattr(value, 'iter_slice') else value[block_start:stop, :]
                        )
                    # print(future_kwargs['data_input'][key])

                # We create a future object with the appropriate name, and then append it to the work manager.
                futures.append(generate_future(self.work_manager, name, eval_block, future_kwargs))

            pi.new_operation('Calculating {}'.format(name), len(futures))

            # Now, we wait to get the result back; we'll store it in the result, and return it.
            for future in self.work_manager.as_completed(futures):
                pi.progress = iblock / step_iter
                future_result = future.get_result(discard=True)

                if dim == 2:
                    for result in future_result:
                        name, iblock, istate, jstate, ci_result = result
                        evolution_dataset[iblock, istate, jstate] = ci_result
                elif dim == 1:
                    for result in future_result:
                        name, iblock, istate, ci_result = result
                        evolution_dataset[iblock, istate] = ci_result

        return evolution_dataset



import itertools
import logging
import re

import numpy as np

import westpa
from westpa.oldtools.aframe import AnalysisMixin

log = logging.getLogger(__name__)


class KineticsAnalysisMixin(AnalysisMixin):
    def __init__(self):
        super().__init__()

        self.dt = None
        self.analysis_initial_bins = None
        self.analysis_final_bins = None

    def add_args(self, parser, upcall=True):
        if upcall:
            try:
                upfunc = super().add_args
            except AttributeError:
                pass
            else:
                upfunc(parser)

        group = parser.add_argument_group('kinetics analysis options')
        group.add_argument(
            '--dt', dest='dt', type=float, default=1.0, help='Assume input data has a time spacing of DT (default: %(default)s).'
        )
        group.add_argument(
            '--initial-bins',
            dest='ibins_string',
            metavar='ILIST',
            help='''Only calculate statistics for transitions starting in bin ILIST.  This may be specified as a
                             comma-separated list of integers or ranges, as in "0,2-4,5,9"''',
        )
        group.add_argument(
            '--final-bins',
            dest='fbins_string',
            metavar='FLIST',
            help='''Only calculate statistics for transitions ending in bin FLIST.  This may be specified as a
                             comma-separated list of integers or ranges, as in "0,2-4,5,9"''',
        )

    def process_args(self, args, upcall=True):
        self.dt = args.dt
        westpa.rc.pstatus('Assuming input data timestep of {:g}'.format(self.dt))

        if args.ibins_string:
            self.analysis_initial_bins = self.parse_bin_range(args.ibins_string)
            westpa.rc.pstatus(
                'Will calculate kinetics data from transitions beginning in the following bins: {!s}'.format(
                    sorted(self.analysis_initial_bins)
                )
            )
        else:
            westpa.rc.pstatus('Will calculate kinetics data from transitions beginning in any bin.')

        if args.fbins_string:
            self.analysis_final_bins = self.parse_bin_range(args.fbins_string)
            westpa.rc.pstatus(
                'Will calculate kinetics data from transitions ending in the following bins: {!s}'.format(
                    sorted(self.analysis_final_bins)
                )
            )
        else:
            westpa.rc.pstatus('Will calculate kinetics data from transitions ending in any bin.')

        if upcall:
            try:
                upfunc = super().process_args
            except AttributeError:
                pass
            else:
                upfunc(args)

    def parse_bin_range(self, range_string):
        try:
            entries = set()
            fields = re.split(r'\s*,\s*', range_string)
            for field in fields:
                if '-' in field:
                    lb, ub = list(map(int, re.split(r'\s*-\s*', field)))
                    entries.update(list(range(lb, ub + 1)))
                else:
                    entries.add(int(field))
        except (ValueError, TypeError):
            raise ValueError('invalid bin range string {!r}'.format(range_string))
        else:
            return entries

    def check_bin_selection(self, n_bins=None):
        '''Check to see that the bin ranges selected by the user conform to the available bins (i.e.,
        bin indices are within the permissible range).  Also assigns the complete bin range if the
        user has not explicitly limited the bins to be considered.'''

        n_bins = n_bins or self.n_bins

        if self.analysis_initial_bins:
            if (np.array(list(self.analysis_initial_bins)) >= n_bins).any():
                raise ValueError('One or more initial bin indices is out of range.')
        else:
            self.analysis_initial_bins = set(range(n_bins))

        if self.analysis_final_bins:
            if (np.array(list(self.analysis_final_bins)) >= n_bins).any():
                raise ValueError('One or more final bin indices is out of range.')
        else:
            self.analysis_final_bins = set(range(n_bins))

    @property
    def selected_bin_pair_iter(self):
        return (tuple(pair) for pair in itertools.product(self.analysis_initial_bins, self.analysis_final_bins))



from collections import namedtuple
from itertools import zip_longest

import numpy as np

import westpa
from westpa.core.kinetics._kinetics import (
    flux_assign,
    pop_assign,
    calc_rates,
    StreamingStats1D,
    StreamingStats2D,
)  # @UnresolvedImport


# Named tuple proxy for StreamingStats class
StreamingStatsTuple = namedtuple('StreamingStatsTuple', ['M1', 'M2', 'n'])


def grouper(n, iterable, fillvalue=None):
    "Collect data into fixed-length chunks or blocks"
    # grouper(3, 'ABCDEFG', 'x') --> ABC DEF Gxx
    args = [iter(iterable)] * n
    return zip_longest(fillvalue=fillvalue, *args)


def tuple2stats(stat_tuple):
    ndims = stat_tuple.M1.ndim
    assert ndims == 1 or ndims == 2

    if ndims == 2:
        stats = StreamingStats2D(stat_tuple.M1.shape)
    elif ndims == 1:
        stats = StreamingStats1D(stat_tuple.M1.shape[0])
    else:
        raise ValueError

    stats.M1 = stat_tuple.M1
    stats.M2 = stat_tuple.M2
    stats.n = stat_tuple.n

    return stats


def process_iter_chunk(bin_mapper, iter_indices, iter_data=None):
    '''Calculate the flux matrices and populations of a set of iterations specified
    by iter_indices. Optionally provide the necessary arrays to perform the calculation
    in iter_data. Otherwise get data from the data_manager directly.
    '''

    data_manager = westpa.rc.get_data_manager()
    system = westpa.rc.get_system_driver()

    # itercount = len(iter_indices)
    nbins = bin_mapper.nbins

    flux_stats = StreamingStats2D((nbins, nbins))
    rate_stats = StreamingStats2D((nbins, nbins))
    pop_stats = StreamingStats1D(nbins)

    nomask1d = np.zeros((nbins,), np.uint8)
    nomask2d = np.zeros((nbins, nbins), np.uint8)
    rate_mask = np.zeros((nbins, nbins), np.uint8)

    flux_matrix = np.zeros((nbins, nbins), np.float64)
    rate_matrix = np.zeros((nbins, nbins), np.float64)
    population_vector = np.zeros((nbins,), np.float64)

    pcoord_len = system.pcoord_len
    assign = bin_mapper.assign

    for iiter, n_iter in enumerate(iter_indices):
        flux_matrix.fill(0.0)
        population_vector.fill(0.0)

        if iter_data:
            iter_group_name = 'iter_{:0{prec}d}'.format(int(n_iter), prec=data_manager.iter_prec)
            iter_group = iter_data[iter_group_name]
        else:
            iter_group = data_manager.get_iter_group(n_iter)

        # first, account for the flux due to recycling
        # we access the hdf5 file directly to avoid nearly 50% overhead of creating a ton of
        # tiny newweightentry objects
        try:
            nwgroup = iter_group['new_weights']
        except KeyError:
            # no new weight data
            pass
        else:
            if iter_data:
                index = None
                weights = nwgroup['weight']
                prev_init_pcoords = nwgroup['prev_init_pcoord']
                new_init_pcoords = nwgroup['new_init_pcoord']
            else:
                index = nwgroup['index'][...]
                weights = index['weight']
                prev_init_pcoords = nwgroup['prev_init_pcoord'][...]
                new_init_pcoords = nwgroup['new_init_pcoord'][...]

            prev_init_assignments = assign(prev_init_pcoords)
            new_init_assignments = assign(new_init_pcoords)

            flux_assign(weights, prev_init_assignments, new_init_assignments, flux_matrix)
            # for (weight,i,j) in izip (weights, prev_init_assignments, new_init_assignments):
            #    flux_matrices[iiter,i,j] += weight
            del index
            del prev_init_pcoords, new_init_pcoords, prev_init_assignments, new_init_assignments, weights

        # iter_group = data_manager.get_iter_group(n_iter)
        if iter_data:
            weights = iter_group['weight']
            initial_pcoords = iter_group['initial_pcoords']
            final_pcoords = iter_group['final_pcoords']
        else:
            weights = iter_group['seg_index']['weight']
            initial_pcoords = iter_group['pcoord'][:, 0]
            final_pcoords = iter_group['pcoord'][:, pcoord_len - 1]

        initial_assignments = assign(initial_pcoords)
        final_assignments = assign(final_pcoords)

        flux_assign(weights, initial_assignments, final_assignments, flux_matrix)
        pop_assign(weights, initial_assignments, population_vector)

        flux_stats.update(flux_matrix, nomask2d)
        pop_stats.update(population_vector, nomask1d)

        calc_rates(flux_matrix, population_vector, rate_matrix, rate_mask)
        rate_stats.update(rate_matrix, rate_mask)

        del weights
        del initial_assignments, final_assignments
        del initial_pcoords, final_pcoords
        del iter_group

    # Create namedtuple proxies for the cython StreamingStats objects
    # since the typed memoryviews class variables do not seem to return
    # cleanly from the zmq workers
    c_flux_stats = StreamingStatsTuple(flux_stats.M1, flux_stats.M2, flux_stats.n)
    c_rate_stats = StreamingStatsTuple(rate_stats.M1, rate_stats.M2, rate_stats.n)
    c_pop_stats = StreamingStatsTuple(pop_stats.M1, pop_stats.M2, pop_stats.n)

    return c_flux_stats, c_rate_stats, c_pop_stats


class RateAverager:
    '''Calculate bin-to-bin kinetic properties (fluxes, rates, populations) at
    1-tau resolution'''

    def __init__(self, bin_mapper, system=None, data_manager=None, work_manager=None):
        self.bin_mapper = bin_mapper
        self.data_manager = data_manager or westpa.rc.get_data_manager()
        self.system = system or westpa.rc.get_system_driver()
        self.work_manager = work_manager or westpa.rc.get_work_manager()

    def extract_data(self, iter_indices):
        '''Extract data from the data_manger and place in dict mirroring the same
        underlying layout.'''

        data = {}
        pcoord_len = self.system.pcoord_len

        for n_iter in iter_indices:
            iter_group_name = 'iter_{:0{prec}d}'.format(int(n_iter), prec=self.data_manager.iter_prec)
            iter_group = self.data_manager.get_iter_group(n_iter)
            di = data[iter_group_name] = {}

            try:
                nwgroup = iter_group['new_weights']
            except KeyError:
                # no new weight data
                pass
            else:
                di_nw = di['new_weights'] = {}
                di_nw['weight'] = nwgroup['index'][...]['weight']
                di_nw['prev_init_pcoord'] = nwgroup['prev_init_pcoord'][...]
                di_nw['new_init_pcoord'] = nwgroup['new_init_pcoord'][...]

            di['weight'] = iter_group['seg_index']['weight']
            di['initial_pcoords'] = iter_group['pcoord'][:, 0]
            di['final_pcoords'] = iter_group['pcoord'][:, pcoord_len - 1]

        return data

    def task_generator(self, iter_start, iter_stop, block_size):
        for iter_block in grouper(block_size, range(iter_start, iter_stop)):
            iter_block = [x for x in iter_block if x is not None]
            iter_data = self.extract_data(iter_block)
            yield (process_iter_chunk, (self.bin_mapper, iter_block), {'iter_data': iter_data})
            del iter_data

    def calculate(self, iter_start=None, iter_stop=None, n_blocks=1, queue_size=1):
        '''Read the HDF5 file and collect flux matrices and population vectors
        for each bin for each iteration in the range [iter_start, iter_stop). Break
        the calculation into n_blocks blocks. If the calculation is broken up into
        more than one block, queue_size specifies the maxmimum number of tasks in
        the work queue.'''

        iter_start = iter_start or 1
        iter_stop = iter_stop or self.data_manager.current_iteration

        itercount = iter_stop - iter_start
        block_size = max(1, itercount // n_blocks)
        nbins = self.bin_mapper.nbins

        if n_blocks == 1:
            flux_stats_t, rate_stats_t, population_stats_t = process_iter_chunk(self.bin_mapper, list(range(iter_start, iter_stop)))

            flux_stats = tuple2stats(flux_stats_t)
            rate_stats = tuple2stats(rate_stats_t)
            population_stats = tuple2stats(population_stats_t)
        else:
            flux_stats = StreamingStats2D((nbins, nbins))
            rate_stats = StreamingStats2D((nbins, nbins))
            population_stats = StreamingStats1D(nbins)

            task_generator = self.task_generator(iter_start, iter_stop, block_size)

            for future in self.work_manager.submit_as_completed(task_generator, queue_size):
                chunk_flux_stats_t, chunk_rate_stats_t, chunk_pop_stats_t = future.get_result()

                chunk_flux_stats = tuple2stats(chunk_flux_stats_t)
                chunk_rate_stats = tuple2stats(chunk_rate_stats_t)
                chunk_pop_stats = tuple2stats(chunk_pop_stats_t)

                # Update statistics with chunked subsets
                flux_stats += chunk_flux_stats
                rate_stats += chunk_rate_stats
                population_stats += chunk_pop_stats

        self.average_flux = flux_stats.mean
        self.stderr_flux = np.nan_to_num(np.sqrt(flux_stats.var) / flux_stats.n)

        self.average_populations = population_stats.mean
        self.stderr_populations = np.nan_to_num(np.sqrt(population_stats.var) / population_stats.n)

        self.average_rate = rate_stats.mean
        self.stderr_rate = np.nan_to_num(np.sqrt(rate_stats.var) / rate_stats.n)

        assert ~np.any(np.isinf(self.stderr_flux))
        assert ~np.any(np.isinf(self.stderr_rate))
        assert ~np.any(np.isinf(self.stderr_populations))


if __name__ == '__main__':
    # Tests this file on the west.h5 data in the current (sim root) directory
    westpa.rc.read_config()
    system = westpa.rc.get_system_driver()
    data_manager = westpa.rc.get_data_manager()
    data_manager.open_backing('r')
    averager = RateAverager(system.bin_mapper)
    averager.calculate()

    print('Population mean and standard error')
    print(averager.average_populations)
    print(averager.stderr_populations)

    print('Flux matrix, mean and standard error')
    print(averager.average_flux)
    print(averager.stderr_flux)

    print('Rate matrix, mean and standard error')
    print(averager.average_rate)
    print(averager.stderr_rate)



'''Miscellaneous routines to help with HDF5 input and output of WEST-related data.'''

import collections
import errno
import getpass
import os
import posixpath
import socket
import sys
import time
import logging
import warnings

import h5py
import numpy as np
from numpy import index_exp
from tables import NaturalNameWarning

from mdtraj import Trajectory, join as join_traj
from mdtraj.utils import in_units_of, import_, ensure_type
from mdtraj.formats import HDF5TrajectoryFile
from mdtraj.formats.hdf5 import _check_mode, Frames

from .trajectory import WESTTrajectory

try:
    import psutil
except ImportError:
    psutil = None

log = logging.getLogger(__name__)
warnings.filterwarnings('ignore', category=NaturalNameWarning)

#
# Constants and globals
#
default_iter_prec = 8

#
# Helper functions
#


def resolve_filepath(path, constructor=h5py.File, cargs=None, ckwargs=None, **addtlkwargs):
    '''Use a combined filesystem and HDF5 path to open an HDF5 file and return the
    appropriate object. Returns (h5file, h5object). The file is opened using
    ``constructor(filename, *cargs, **ckwargs)``.'''

    cargs = cargs or ()
    ckwargs = ckwargs or {}
    ckwargs.update(addtlkwargs)
    objpieces = collections.deque()
    path = posixpath.normpath(path)

    filepieces = path.split('/')
    while filepieces:
        testpath = '/'.join(filepieces)
        if not testpath:
            filepieces.pop()
            continue
        try:
            h5file = constructor(testpath, *cargs, **ckwargs)
        except IOError:
            objpieces.appendleft(filepieces.pop())
            continue
        else:
            return (h5file, h5file['/'.join([''] + list(objpieces)) if objpieces else '/'])
    else:
        # We don't provide a filename, because we're not sure where the filename stops
        # and the HDF5 path begins.
        raise IOError(errno.ENOENT, os.strerror(errno.ENOENT))


def calc_chunksize(shape, dtype, max_chunksize=262144):
    '''Calculate a chunk size for HDF5 data, anticipating that access will slice
    along lower dimensions sooner than higher dimensions.'''

    chunk_shape = list(shape)
    dtype = np.dtype(dtype)
    for idim in range(len(shape)):
        chunk_nbytes = np.multiply.reduce(chunk_shape) * dtype.itemsize
        while chunk_shape[idim] > 1 and chunk_nbytes > max_chunksize:
            chunk_shape[idim] >>= 1  # divide by 2
            chunk_nbytes = np.multiply.reduce(chunk_shape) * dtype.itemsize

        if chunk_nbytes <= max_chunksize:
            break

    chunk_shape = tuple(chunk_shape)
    return chunk_shape


def tostr(b):
    '''Convert a nonstandard string object ``b`` to str with the handling of the
    case where ``b`` is bytes.'''

    if b is None:
        return None
    elif isinstance(b, bytes):
        return b.decode('utf-8')
    else:
        return str(b)


def is_within_directory(directory, target):
    abs_directory = os.path.abspath(directory)
    abs_target = os.path.abspath(target)

    prefix = os.path.commonprefix([abs_directory, abs_target])

    return prefix == abs_directory


def safe_extract(tar, path=".", members=None, *, numeric_owner=False):
    for member in tar.getmembers():
        member_path = os.path.join(path, member.name)
        if not is_within_directory(path, member_path):
            raise Exception("Attempted Path Traversal in Tar File")

    tar.extractall(path, members, numeric_owner=numeric_owner)


#
# Group and dataset manipulation functions
#


def create_hdf5_group(parent_group, groupname, replace=False, creating_program=None):
    '''Create (or delete and recreate) and HDF5 group named ``groupname`` within
    the enclosing Group (object) ``parent_group``. If ``replace`` is True, then
    the group is replaced if present; if False, then an error is raised if the
    group is present. After the group is created, HDF5 attributes are set using
    `stamp_creator_data`.
    '''

    if replace:
        try:
            del parent_group[groupname]
        except KeyError:
            pass

    newgroup = parent_group.create_group(groupname)
    stamp_creator_data(newgroup)
    return newgroup


#
# Group and dataset labeling functions
#


def stamp_creator_data(h5group, creating_program=None):
    '''Mark the following on the HDF5 group ``h5group``:

      :creation_program:   The name of the program that created the group
      :creation_user:      The username of the user who created the group
      :creation_hostname:  The hostname of the machine on which the group was created
      :creation_time:      The date and time at which the group was created, in the
                           current locale.
      :creation_unix_time: The Unix time (seconds from the epoch, UTC) at which the
                           group was created.

    This is meant to facilitate tracking the flow of data, but should not be considered
    a secure paper trail (after all, anyone with write access to the HDF5 file can modify
    these attributes).
    '''
    now = time.time()
    attrs = h5group.attrs

    attrs['creation_program'] = creating_program or sys.argv[0] or 'unknown program'
    attrs['creation_user'] = getpass.getuser()
    attrs['creation_hostname'] = socket.gethostname()
    attrs['creation_unix_time'] = now
    attrs['creation_time'] = time.strftime('%c', time.localtime(now))


def get_creator_data(h5group):
    '''Read back creator data as written by ``stamp_creator_data``, returning a dictionary with
    keys as described for ``stamp_creator_data``. Missing fields are denoted with None.
    The ``creation_time`` field is returned as a string.'''
    attrs = h5group.attrs
    d = dict()
    for attr in ['creation_program', 'creation_user', 'creation_hostname', 'creation_unix_time', 'creation_time']:
        d[attr] = attrs.get(attr)
    return d


def load_west(filename):
    """Load WESTPA trajectory files from disk.

    Parameters
    ----------
    filename : str
        String filename of HDF Trajectory file.
    """

    with h5py.File(filename, 'r') as f:
        iter_group_template = 'iter_{1:0{0}d}'
        iter_prec = f.attrs['west_iter_prec']
        trajectories = []
        n = 0

        iter_group_name = iter_group_template.format(iter_prec, n)
        for iter_group_name in f['iterations']:
            iter_group = f['iterations/' + iter_group_name]

            if 'trajectories' in iter_group:
                traj_link = iter_group['trajectories']
                traj_filename = traj_link.file.filename

                with WESTIterationFile(traj_filename) as traj_file:
                    traj = traj_file.read_as_traj()
            else:
                # TODO: [HDF5] allow initializing trajectory without coordinates
                raise ValueError("Missing trajectories for iteration %d" % n)

            # pcoord is required
            if 'pcoord' not in iter_group:
                raise ValueError("Missing pcoords for iteration %d" % n)

            raw_pcoord = iter_group['pcoord'][:]
            if raw_pcoord.ndim != 3:
                log.warning('pcoord is expected to be a 3-d ndarray instead of {}-d'.format(raw_pcoord.ndim))
                continue
            # ignore the first frame of each segment
            if raw_pcoord.shape[1] == traj.n_frames + 1:
                raw_pcoord = raw_pcoord[:, 1:, :]
            elif raw_pcoord.shape[1] == traj.n_frames:
                raw_pcoord = raw_pcoord[:, :, :]
            else:
                raise ValueError(
                    "Inconsistent number of pcoords (%d) and frames (%d) for iteration %d" % (raw_pcoord.shape[1], traj.n_frames, n)
                )

            pcoords = np.concatenate(raw_pcoord, axis=0)
            n_frames = raw_pcoord.shape[1]

            if 'seg_index' in iter_group:
                raw_pid = iter_group['seg_index']['parent_id'][:]

                if np.any(raw_pid < 0):
                    init_basis_ids = iter_group['ibstates']['istate_index']['basis_state_id'][:]
                    init_ids = -(raw_pid[raw_pid < 0] + 1)
                    raw_pid[raw_pid < 0] = [init_basis_ids[iid] for iid in init_ids]
                parent_ids = raw_pid.repeat(n_frames, axis=0)
            else:
                parent_ids = None

            traj.pcoords = pcoords
            traj.parent_ids = parent_ids
            trajectories.append(traj)

            n += 1
            iter_group_name = iter_group_template.format(iter_prec, n)

    west_traj = join_traj(trajectories)

    return west_traj


###
# Iteration range metadata
###
def stamp_iter_range(h5object, start_iter, stop_iter):
    '''Mark that the HDF5 object ``h5object`` (dataset or group) contains data from iterations
    start_iter <= n_iter < stop_iter.'''
    h5object.attrs['iter_start'] = start_iter
    h5object.attrs['iter_stop'] = stop_iter


def get_iter_range(h5object):
    '''Read back iteration range data written by ``stamp_iter_range``'''
    return int(h5object.attrs['iter_start']), int(h5object.attrs['iter_stop'])


def stamp_iter_step(h5group, iter_step):
    '''Mark that the HDF5 object ``h5object`` (dataset or group) contains data with an
    iteration step (stride) of iter_step).'''
    h5group.attrs['iter_step'] = iter_step


def get_iter_step(h5group):
    '''Read back iteration step (stride) written by ``stamp_iter_step``'''
    return int(h5group.attrs['iter_step'])


def check_iter_range_least(h5object, iter_start, iter_stop):
    '''Return True if the iteration range [iter_start, iter_stop) is
    the same as or entirely contained within the iteration range stored
    on ``h5object``.'''
    obj_iter_start, obj_iter_stop = get_iter_range(h5object)
    return obj_iter_start <= iter_start and obj_iter_stop >= iter_stop


def check_iter_range_equal(h5object, iter_start, iter_stop):
    '''Return True if the iteration range [iter_start, iter_stop) is
    the same as the iteration range stored on ``h5object``.'''
    obj_iter_start, obj_iter_stop = get_iter_range(h5object)
    return obj_iter_start == iter_start and obj_iter_stop == iter_stop


def get_iteration_entry(h5object, n_iter):
    '''Create a slice for data corresponding to iteration ``n_iter`` in ``h5object``.'''
    obj_iter_start, obj_iter_stop = get_iter_range(h5object)
    if n_iter < obj_iter_start or n_iter >= obj_iter_stop:
        raise IndexError('data for iteration {} not available in dataset {!r}'.format(n_iter, h5object))
    return np.index_exp[n_iter - obj_iter_start]


def get_iteration_slice(h5object, iter_start, iter_stop=None, iter_stride=None):
    '''Create a slice for data corresponding to iterations [iter_start,iter_stop),
    with stride iter_step, in the given ``h5object``.'''
    obj_iter_start, obj_iter_stop = get_iter_range(h5object)

    if iter_stop is None:
        iter_stop = iter_start + 1
    if iter_stride is None:
        iter_stride = 1

    if iter_start < obj_iter_start:
        raise IndexError('data for iteration {} not available in dataset {!r}'.format(iter_start, h5object))
    elif iter_start > obj_iter_stop:
        raise IndexError('data for iteration {} not available in dataset {!r}'.format(iter_stop, h5object))

    start_index = iter_start - obj_iter_start
    stop_index = iter_stop - obj_iter_start
    return np.index_exp[start_index:stop_index:iter_stride]


###
# Axis label metadata
###
def label_axes(h5object, labels, units=None):
    '''Stamp the given HDF5 object with axis labels. This stores the axis labels
    in an array of strings in an attribute called ``axis_labels`` on the given
    object. ``units`` if provided is a corresponding list of units.'''

    if len(labels) != len(h5object.shape):
        raise ValueError('number of axes and number of labels do not match')

    if units is None:
        units = []

    if len(units) and len(units) != len(labels):
        raise ValueError('number of units labels does not match number of axes')

    h5object.attrs['axis_labels'] = np.array([np.string_(i) for i in labels])

    if len(units):
        h5object.attrs['axis_units'] = np.array([np.string_(i) for i in units])


NotGiven = object()


def _get_one_attr(h5object, namelist, default=NotGiven):
    attrs = dict(h5object.attrs)
    for name in namelist:
        try:
            return attrs[name]
        except KeyError:
            pass
    else:
        if default is NotGiven:
            raise KeyError('no such key')
        else:
            return default


class WESTPAH5File(h5py.File):
    '''Generalized input/output for WESTPA simulation (or analysis) data.'''

    default_iter_prec = 8
    _this_fileformat_version = 8

    def __init__(self, *args, **kwargs):
        # These values are used for creating files or reading files where this
        # data is not stored. Otherwise, values stored as attributes on the root
        # group are used instead.
        arg_iter_prec = kwargs.pop('westpa_iter_prec', self.default_iter_prec)
        arg_fileformat_version = kwargs.pop('westpa_fileformat_version', self._this_fileformat_version)
        arg_creating_program = kwargs.pop('creating_program', None)

        # Initialize h5py file
        super().__init__(*args, **kwargs)

        # Try to get iteration precision and I/O class version
        h5file_iter_prec = _get_one_attr(self, ['westpa_iter_prec', 'west_iter_prec', 'wemd_iter_prec'], None)
        h5file_fileformat_version = _get_one_attr(
            self, ['westpa_fileformat_version', 'west_file_format_version', 'wemd_file_format_version'], None
        )

        self.iter_prec = h5file_iter_prec if h5file_iter_prec is not None else arg_iter_prec
        self.fileformat_version = h5file_fileformat_version if h5file_fileformat_version is not None else arg_fileformat_version

        # Ensure that file format attributes are stored, if the file is writable
        if self.mode == 'r+':
            self.attrs['westpa_iter_prec'] = self.iter_prec
            self.attrs['westpa_fileformat_version'] = self.fileformat_version
            if arg_creating_program:
                stamp_creator_data(self, creating_program=arg_creating_program)

    # Helper function to automatically replace a group, if it exists.
    # Should really only be called when one is certain a dataset should be blown away.
    def replace_dataset(self, *args, **kwargs):
        try:
            del self[args[0]]
        except Exception:
            pass
        try:
            del self[kwargs['name']]
        except Exception:
            pass

        return self.create_dataset(*args, **kwargs)

    # Iteration groups

    def iter_object_name(self, n_iter, prefix='', suffix=''):
        '''Return a properly-formatted per-iteration name for iteration
        ``n_iter``. (This is used in create/require/get_iter_group, but may
        also be useful for naming datasets on a per-iteration basis.)'''
        return '{prefix}iter_{n_iter:0{prec}d}{suffix}'.format(n_iter=n_iter, prefix=prefix, suffix=suffix, prec=self.iter_prec)

    def create_iter_group(self, n_iter, group=None):
        '''Create a per-iteration data storage group for iteration number ``n_iter``
        in the group ``group`` (which is '/iterations' by default).'''

        if group is None:
            group = self.require_group('/iterations')
        return group.create_group(self.iter_object_name(n_iter))

    def require_iter_group(self, n_iter, group=None):
        '''Ensure that a per-iteration data storage group for iteration number ``n_iter``
        is available in the group ``group`` (which is '/iterations' by default).'''
        if group is None:
            group = self.require_group('/iterations')
        return group.require_group(self.iter_object_name(n_iter))

    def get_iter_group(self, n_iter, group=None):
        '''Get the per-iteration data group for iteration number ``n_iter`` from within
        the group ``group`` ('/iterations' by default).'''
        if group is None:
            group = self['/iterations']
        return group[self.iter_object_name(n_iter)]


class WESTIterationFile(HDF5TrajectoryFile):
    def __init__(self, file, mode='r', force_overwrite=True, compression='zlib', link=None):
        if isinstance(file, str):
            super(WESTIterationFile, self).__init__(file, mode, force_overwrite, compression)
        else:
            try:
                self._init_from_handle(file)
            except AttributeError:
                raise ValueError('unknown input type: %s' % str(type(file)))

    def _init_from_handle(self, handle):
        self._handle = handle
        self._open = handle.isopen != 0
        self.mode = mode = handle.mode  # the mode in which the file was opened?

        if mode not in ['r', 'w', 'a']:
            raise ValueError("mode must be one of ['r', 'w', 'a']")

        # import tables
        self.tables = import_('tables')

        if mode == 'w':
            # what frame are we currently reading or writing at?
            self._frame_index = 0
            # do we need to write the header information?
            self._needs_initialization = True

        elif mode == 'a':
            try:
                self._frame_index = len(self._handle.root.coordinates)
                self._needs_initialization = False
            except self.tables.NoSuchNodeError:
                self._frame_index = 0
                self._needs_initialization = True
        elif mode == 'r':
            self._frame_index = 0
            self._needs_initialization = False

    def read(self, frame_indices=None, atom_indices=None):
        _check_mode(self.mode, ('r',))

        if frame_indices is None:
            frame_slice = slice(None)
            self._frame_index += frame_slice.stop - frame_slice.start
        else:
            frame_slice = ensure_type(frame_indices, dtype=int, ndim=1, name='frame_indices', warn_on_cast=False)
            if not np.all(frame_slice < self._handle.root.coordinates.shape[0]):
                raise ValueError(
                    'As a zero-based index, the entries in '
                    'frame_slice must all be less than the number of frames '
                    'in the trajectory, %d' % self._handle.root.coordinates.shape[0]
                )
            if not np.all(frame_slice >= 0):
                raise ValueError('The entries in frame_indices must be greater ' 'than or equal to zero')
            self._frame_index += frame_slice[-1] - frame_slice[0]

        if atom_indices is None:
            # get all of the atoms
            atom_slice = slice(None)
        else:
            atom_slice = ensure_type(atom_indices, dtype=int, ndim=1, name='atom_indices', warn_on_cast=False)
            if not np.all(atom_slice < self._handle.root.coordinates.shape[1]):
                raise ValueError(
                    'As a zero-based index, the entries in '
                    'atom_indices must all be less than the number of atoms '
                    'in the trajectory, %d' % self._handle.root.coordinates.shape[1]
                )
            if not np.all(atom_slice >= 0):
                raise ValueError('The entries in atom_indices must be greater ' 'than or equal to zero')

        def get_item(node, key):
            if not isinstance(key, tuple):
                return node.__getitem__(key)

            n_list_like = 0
            new_keys = []
            for item in key:
                if not isinstance(item, slice):
                    try:
                        d = np.diff(item)
                        if len(d) == 0:
                            item = item[0]
                        elif np.all(d == d[0]):
                            item = slice(item[0], item[-1] + d[0], d[0])
                        else:
                            n_list_like += 1
                    except Exception:
                        n_list_like += 1
                new_keys.append(item)
            new_keys = tuple(new_keys)

            if n_list_like <= 1:
                return node.__getitem__(new_keys)

            data = node
            for i, item in enumerate(new_keys):
                dkey = [slice(None)] * len(key)
                dkey[i] = item
                dkey = tuple(dkey)
                data = data.__getitem__(dkey)

            return data

        def get_field(name, slice, out_units, can_be_none=True):
            try:
                node = self._get_node(where='/', name=name)
                data = get_item(node, slice)
                in_units = node.attrs.units
                if not isinstance(in_units, str):
                    in_units = in_units.decode()
                data = in_units_of(data, in_units, out_units)
                return data
            except self.tables.NoSuchNodeError:
                if can_be_none:
                    return None
                raise

        frames = Frames(
            coordinates=get_field('coordinates', (frame_slice, atom_slice, slice(None)), out_units='nanometers', can_be_none=False),
            time=get_field('time', frame_slice, out_units='picoseconds'),
            cell_lengths=get_field('cell_lengths', (frame_slice, slice(None)), out_units='nanometers'),
            cell_angles=get_field('cell_angles', (frame_slice, slice(None)), out_units='degrees'),
            velocities=get_field('velocities', (frame_slice, atom_slice, slice(None)), out_units='nanometers/picosecond'),
            kineticEnergy=get_field('kineticEnergy', frame_slice, out_units='kilojoules_per_mole'),
            potentialEnergy=get_field('potentialEnergy', frame_slice, out_units='kilojoules_per_mole'),
            temperature=get_field('temperature', frame_slice, out_units='kelvin'),
            alchemicalLambda=get_field('lambda', frame_slice, out_units='dimensionless'),
        )

        return frames

    def _has_node(self, where, name):
        try:
            self._get_node(where, name=name)
        except self.tables.NoSuchNodeError:
            return False

        return True

    def has_topology(self):
        return self._has_node('/', 'topology')

    def has_pointer(self):
        return self._has_node('/', 'pointer')

    def has_restart(self, segment):
        return self._has_node('/restart', '%d_%d' % (segment.n_iter, segment.seg_id))

    def write_data(self, where, name, data):
        node = self._get_node(where=where, name=name)
        node.append(data)

    def read_data(self, where, name):
        node = self._get_node(where=where, name=name)
        return node.read()

    def read_as_traj(self, iteration=None, segment=None, atom_indices=None):
        _check_mode(self.mode, ('r',))

        pnode = self._get_node(where='/', name='pointer')

        iter_labels = pnode[:, 0]
        seg_labels = pnode[:, 1]

        if iteration is None and segment is None:
            frame_indices = slice(None)
        elif isinstance(iteration, (np.integer, int)) and isinstance(segment, (np.integer, int)):
            frame_torf = np.logical_and(iter_labels == iteration, seg_labels == segment)
            frame_indices = np.arange(len(iter_labels))[frame_torf]
        else:
            raise ValueError("iteration and segment must be integers and provided at the same time")

        if len(frame_indices) == 0:
            raise ValueError(f"no frame was selected: iteration={iteration}, segment={segment}, atom_indices={atom_indices}")

        iter_labels = iter_labels[frame_indices]
        seg_labels = seg_labels[frame_indices]

        topology = self.topology
        if atom_indices is not None:
            topology = topology.subset(atom_indices)

        data = self.read(frame_indices=frame_indices, atom_indices=atom_indices)
        if len(data) == 0:
            return Trajectory(xyz=np.zeros((0, topology.n_atoms, 3)), topology=topology)

        in_units_of(data.coordinates, self.distance_unit, Trajectory._distance_unit, inplace=True)
        in_units_of(data.cell_lengths, self.distance_unit, Trajectory._distance_unit, inplace=True)

        return WESTTrajectory(
            data.coordinates,
            topology=topology,
            time=data.time,
            unitcell_lengths=data.cell_lengths,
            unitcell_angles=data.cell_angles,
            iter_labels=iter_labels,
            seg_labels=seg_labels,
            pcoords=None,
        )

    def read_restart(self, segment):
        if self.has_restart(segment):
            data = self.read_data('/restart/%d_%d' % (segment.n_iter, segment.seg_id), 'data')
            segment.data['iterh5/restart'] = data
        else:
            raise ValueError('no restart data available for {}'.format(str(segment)))

    def write_segment(self, segment, pop=False):
        n_iter = segment.n_iter

        self.root._v_attrs['n_iter'] = n_iter

        if pop:
            get_data = segment.data.pop
        else:
            get_data = segment.data.get

        traj = get_data('iterh5/trajectory', None)
        restart = get_data('iterh5/restart', None)
        slog = get_data('iterh5/log', None)

        if traj is not None:
            # create trajectory object
            traj = WESTTrajectory(traj, iter_labels=n_iter, seg_labels=segment.seg_id)
            if traj.n_frames == 0:
                # we may consider logging warnings instead throwing errors for later.
                # right now this is good for debugging purposes
                raise ValueError('no trajectory data present for %s' % repr(segment))

            if n_iter == 0:
                base_time = 0
            else:
                iter_duration = traj.time[-1] - traj.time[0]
                base_time = iter_duration * (n_iter - 1)

            traj.time -= traj.time[0]
            traj.time += base_time

            # pointers
            if not self.has_pointer():
                self._create_earray('/', name='pointer', atom=self.tables.Int64Atom(), shape=(0, 2))

            iter_idx = traj.iter_labels
            seg_idx = traj.seg_labels

            pointers = np.stack((iter_idx, seg_idx)).T

            self.write_data('/', 'pointer', pointers)

            # trajectory
            self.write(
                coordinates=in_units_of(traj.xyz, Trajectory._distance_unit, self.distance_unit),
                time=traj.time,
                cell_lengths=in_units_of(traj.unitcell_lengths, Trajectory._distance_unit, self.distance_unit),
                cell_angles=traj.unitcell_angles,
            )

            # topology
            if self.mode == 'a':
                if not self.has_topology():
                    self.topology = traj.topology
            elif self.mode == 'w':
                self.topology = traj.topology

        # restart
        if restart is not None:
            if self.has_restart(segment):
                self._remove_node('/restart', name='%d_%d' % (segment.n_iter, segment.seg_id), recursive=True)

            self._create_array(
                '/restart/%d_%d' % (segment.n_iter, segment.seg_id),
                name='data',
                atom=self.tables.StringAtom(itemsize=len(restart)),
                obj=restart,
                createparents=True,
            )

        if slog is not None:
            if self._has_node('/log', str(segment.seg_id)):
                self._remove_node('/log', name=str(segment.seg_id), recursive=True)

            self._create_array(
                '/log/%d_%d' % (segment.n_iter, segment.seg_id),
                name='data',
                atom=self.tables.StringAtom(itemsize=len(slog)),
                obj=slog,
                createparents=True,
            )

    @property
    def _create_group(self):
        if self.tables.__version__ >= '3.0.0':
            return self._handle.create_group
        return self._handle.createGroup

    @property
    def _create_array(self):
        if self.tables.__version__ >= '3.0.0':
            return self._handle.create_array
        return self._handle.createArray

    @property
    def _remove_node(self):
        if self.tables.__version__ >= '3.0.0':
            return self._handle.remove_node
        return self._handle.removeNode


### Generalized WE dataset access classes


class DSSpec:
    '''Generalized WE dataset access'''

    def get_iter_data(self, n_iter, seg_slice=index_exp[:]):
        raise NotImplementedError

    def get_segment_data(self, n_iter, seg_id):
        return self.get_iter_data(n_iter)[seg_id]

    def __getstate__(self):
        d = dict(self.__dict__)
        if '_h5file' in d:
            d['_h5file'] = None
        return d

    def __setstate__(self, state):
        self.__dict__.update(state)


class FileLinkedDSSpec(DSSpec):
    '''Provide facilities for accessing WESTPA HDF5 files, including auto-opening and the ability
    to pickle references to such files for transmission (through, e.g., the work manager), provided
    that the HDF5 file can be accessed by the same path on both the sender and receiver.'''

    def __init__(self, h5file_or_name):
        self._h5file = None
        self._h5filename = None

        try:
            self._h5filename = os.path.abspath(h5file_or_name.filename)
        except AttributeError:
            self._h5filename = h5file_or_name
            self._h5file = None
        else:
            self._h5file = h5file_or_name

    @property
    def h5file(self):
        '''Lazily open HDF5 file. This is required because allowing an open HDF5
        file to cross a fork() boundary generally corrupts the internal state of
        the HDF5 library.'''
        if self._h5file is None:
            self._h5file = WESTPAH5File(self._h5filename, 'r')
        return self._h5file


class SingleDSSpec(FileLinkedDSSpec):
    @classmethod
    def from_string(cls, dsspec_string, default_h5file):
        alias = None

        h5file = default_h5file
        fields = dsspec_string.split(',')
        dsname = fields[0]
        slice = None

        for field in (field.strip() for field in fields[1:]):
            k, v = field.split('=')
            k = k.lower()
            if k == 'alias':
                alias = v
            elif k == 'slice':
                try:
                    slice = eval('np.index_exp' + v)
                except SyntaxError:
                    raise SyntaxError('invalid index expression {!r}'.format(v))
            elif k == 'file':
                h5file = v
            else:
                raise ValueError('invalid dataset option {!r}'.format(k))

        return cls(h5file, dsname, alias, slice)

    def __init__(self, h5file_or_name, dsname, alias=None, slice=None):
        FileLinkedDSSpec.__init__(self, h5file_or_name)
        self.dsname = dsname
        self.alias = alias or dsname
        self.slice = np.index_exp[slice] if slice else None


class SingleIterDSSpec(SingleDSSpec):
    def get_iter_data(self, n_iter, seg_slice=index_exp[:]):
        if self.slice:
            return self.h5file.get_iter_group(n_iter)[self.dsname][seg_slice + self.slice]
        else:
            return self.h5file.get_iter_group(n_iter)[self.dsname][seg_slice]


class SingleSegmentDSSpec(SingleDSSpec):
    def get_iter_data(self, n_iter, seg_slice=index_exp[:]):
        if self.slice:
            return self.h5file.get_iter_group(n_iter)[self.dsname][seg_slice + index_exp[:] + self.slice]
        else:
            return self.h5file.get_iter_group(n_iter)[self.dsname][seg_slice]

    def get_segment_data(self, n_iter, seg_id):
        if self.slice:
            return self.h5file.get_iter_group(n_iter)[np.index_exp[seg_id, :] + self.slice]
        else:
            return self.h5file.get_iter_group(n_iter)[seg_id]


class FnDSSpec(FileLinkedDSSpec):
    def __init__(self, h5file_or_name, fn):
        FileLinkedDSSpec.__init__(self, h5file_or_name)
        self.fn = fn

    def get_iter_data(self, n_iter, seg_slice=index_exp[:]):
        return self.fn(n_iter, self.h5file.get_iter_group(n_iter))[seg_slice]


class MultiDSSpec(DSSpec):
    def __init__(self, dsspecs):
        self.dsspecs = dsspecs

    def get_iter_data(self, n_iter, seg_slice=index_exp[:]):
        datasets = [dsspec.get_iter_data(n_iter) for dsspec in self.dsspecs]

        ncols = 0
        nsegs = None
        npts = None
        for iset, dset in enumerate(datasets):
            if nsegs is None:
                nsegs = dset.shape[0]
            elif dset.shape[0] != nsegs:
                raise TypeError('dataset {} has incorrect first dimension (number of segments)'.format(self.dsspecs[iset]))
            if npts is None:
                npts = dset.shape[1]
            elif dset.shape[1] != npts:
                raise TypeError('dataset {} has incorrect second dimension (number of time points)'.format(self.dsspecs[iset]))

            if dset.ndim < 2:
                # scalar per segment or scalar per iteration
                raise TypeError('dataset {} has too few dimensions'.format(self.dsspecs[iset]))
            elif dset.ndim > 3:
                # array per timepoint
                raise TypeError('dataset {} has too many dimensions'.format(self.dsspecs[iset]))
            elif dset.ndim == 2:
                # scalar per timepoint
                ncols += 1
            else:
                # vector per timepoint
                ncols += dset.shape[-1]

        output_dtype = np.result_type(*[ds.dtype for ds in datasets])
        output_array = np.empty((nsegs, npts, ncols), dtype=output_dtype)

        ocol = 0
        for iset, dset in enumerate(datasets):
            if dset.ndim == 2:
                output_array[:, :, ocol] = dset[...]
                ocol += 1
            elif dset.ndim == 3:
                output_array[:, :, ocol : (ocol + dset.shape[-1])] = dset[...]
                ocol += dset.shape[-1]

        return output_array[seg_slice]


class IterBlockedDataset:
    @classmethod
    def empty_like(cls, blocked_dataset):
        source = blocked_dataset.data if blocked_dataset.data is not None else blocked_dataset.dataset

        newbds = cls(
            np.empty(source.shape, source.dtype),
            attrs={'iter_start': blocked_dataset.iter_start, 'iter_stop': blocked_dataset.iter_stop},
        )
        return newbds

    def __init__(self, dataset_or_array, attrs=None):
        try:
            dataset_or_array.attrs
        except AttributeError:
            self.dataset = None
            self.data = dataset_or_array
            if attrs is None:
                raise ValueError('attribute dictionary containing iteration bounds must be provided')
            self.iter_shape = self.data.shape[1:]
            self.dtype = self.data.dtype
        else:
            self.dataset = dataset_or_array
            attrs = self.dataset.attrs
            self.data = None
            self.iter_shape = self.dataset.shape[1:]
            self.dtype = self.dataset.dtype

        self.iter_start = attrs['iter_start']
        self.iter_stop = attrs['iter_stop']

    def cache_data(self, max_size=None):
        '''Cache this dataset in RAM. If ``max_size`` is given, then only cache if the entire dataset
        fits in ``max_size`` bytes. If ``max_size`` is the string 'available', then only cache if
        the entire dataset fits in available RAM, as defined by the ``psutil`` module.'''

        if max_size is not None:
            dssize = self.dtype.itemsize * np.multiply.reduce(self.dataset.shape)
            if max_size == 'available' and psutil is not None:
                avail_bytes = psutil.virtual_memory().available
                if dssize > avail_bytes:
                    return
            elif isinstance(max_size, str):
                return
            else:
                if dssize > max_size:
                    return
        if self.dataset is not None:
            if self.data is None:
                self.data = self.dataset[...]

    def drop_cache(self):
        if self.dataset is not None:
            del self.data
            self.data = None

    def iter_entry(self, n_iter):
        if n_iter < self.iter_start:
            raise IndexError('requested iteration {} less than first stored iteration {}'.format(n_iter, self.iter_start))

        source = self.data if self.data is not None else self.dataset
        return source[n_iter - self.iter_start]

    def iter_slice(self, start=None, stop=None):
        start = start or self.iter_start
        stop = stop or self.iter_stop
        step = 1  # strided retrieval not implemented yet

        # if step % self.iter_step > 0:
        #    raise TypeError('dataset {!r} stored with stride {} cannot be accessed with stride {}'
        #                    .format(self.dataset, self.iter_step, step))
        if start < self.iter_start:
            raise IndexError('requested start {} less than stored start {}'.format(start, self.iter_start))
        elif stop > self.iter_stop:
            stop = self.iter_stop

        source = self.data if self.data is not None else self.dataset
        return source[start - self.iter_start : stop - self.iter_start : step]
